// Surface3D.cpp: implementation of the CSurface3D class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Surface3D.h"
#include "..\globals.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSurface3D::CSurface3D()
{
	vertexRefQty = 0;
	vertexRefs = NULL;

	alpha = 255;
	lightmapID = -1;
	decalListID = -1;

	normal.SetVec(0.0f,0.0f,0.0f);
	planeK = 0.0f;

	texture = NULL;
	opposingSurface = NULL;

	texCoords = NULL;
	texCoordsLM = NULL;

	segUnits = NULL;
	segNormals = NULL;
	segLengths = NULL;

	useColor = false;

	visible = true;
	mirror = false;
	floor = false;

	dynLMID = -1;
	dynLMXf = 0.0f;
	dynLMYf = 0.0f;
	dynLMX = 0;
	dynLMY = 0;
}

CSurface3D::~CSurface3D()
{
	Destroy();
}

void CSurface3D::Destroy(void)
{
	// Destroy texcoord arrays, vertexRef, and segment arrays
	// If this gets moved to a Destroy function, set the pointers
	//   to NULL!
	if (vertexRefs != NULL) delete [] vertexRefs;
	if (segNormals != NULL) delete [] segNormals;
	if (segUnits != NULL) delete [] segUnits;
	if (segLengths != NULL) delete [] segLengths;
	if (texCoords != NULL) delete [] texCoords;
	if (texCoordsLM != NULL) delete [] texCoordsLM;

	vertexRefs = NULL;
	segNormals = NULL;
	segUnits = NULL;
	segLengths = NULL;
	texCoords = NULL;
	texCoordsLM = NULL;

	vertexRefQty = 0;
}

void CSurface3D::Initialize(unsigned p_vertexQty, CAPITexture *p_texture)
{
	if (p_vertexQty > 0)
	{
		vertexRefQty = p_vertexQty;
		vertexRefs = new int[p_vertexQty];
		segNormals = new CVector3D[p_vertexQty];
		segUnits = new CVector3D[p_vertexQty];
		segLengths = new float[p_vertexQty];

		// Since a surface sould have its texture changed or cleared at any time,
		//   might as well allocate space
		texCoords = new CTexCoord2f[p_vertexQty];
		texCoordsLM = new CTexCoord2f[p_vertexQty];
	}

	texture = p_texture;
}

void CSurface3D::CalculateSurfaceData(CModelVertex3D *p_vertices)
{
    CVector3D v1;
    unsigned i;
    
    if (vertexRefQty < 2)
	{
        normal.SetVec(0, 0, 0);
		return;
	}
    
    if (vertexRefQty == 2)
	{
        // If the 'surface' is on just the
        //   x,y  plane, we can determine one...
        // But it is treated as Ortho, so the normal's
        //   rotation is opposite
        // (imagine the line being part of a clockwise room
        //   defined in windows coordinates)
        if ((p_vertices[vertexRefs[0]].vertex.z == 0) && (p_vertices[vertexRefs[1]].vertex.z == 0))
		{
            v1 = p_vertices[vertexRefs[1]].vertex - p_vertices[vertexRefs[0]].vertex;
            normal.SetVec(-v1.y, v1.x, 0.0f);
            normal.Unitize();
			planeK = -(p_vertices[vertexRefs[0]].vertex * normal);
		}
        else
		{
            normal.SetVec(0, 0, 0);
			planeK = 0.0;
		}
        
        v1 = p_vertices[vertexRefs[1]].vertex - p_vertices[vertexRefs[0]].vertex;
        segLengths[0] = v1.MgnTud();
	}
    else
	{
        // Make sure we get a good one, can't make a normal
        //   off parallel segments
        for (i=0; i<=vertexRefQty-3; i++)
		{
            v1 = (p_vertices[vertexRefs[i]].vertex - p_vertices[vertexRefs[i + 1]].vertex).CrsProd(
                    p_vertices[vertexRefs[i + 2]].vertex - p_vertices[vertexRefs[i + 1]].vertex);
            if (v1.MgnTudSqr() > 0.00001f)
			{
                v1.Unitize();
                normal = v1;
				planeK = -(p_vertices[vertexRefs[0]].vertex * normal);
                i = vertexRefQty - 3;
			}
            //Debug.Print v1.x, v1.y, v1.z
        }
        //if (normal.MgnTudSqr() == 0)
        //    MsgBox "Warning - zero normal"
        
        for (i=0; i<vertexRefQty; i++)
		{
            if (i < vertexRefQty - 1)
                v1 = p_vertices[vertexRefs[i + 1]].vertex - p_vertices[vertexRefs[i]].vertex;
            else
                v1 = p_vertices[vertexRefs[0]].vertex - p_vertices[vertexRefs[i]].vertex;
            
            segLengths[i] = v1.MgnTud();
            v1.Unitize();
            segUnits[i] = v1;
            segNormals[i] = segUnits[i].CrsProd(normal);
		}
    }
}

void CSurface3D::SetVertexRef(unsigned p_index, int p_vertexRefID)
{
	SetVertexRef(p_index, p_vertexRefID, 0,0);
}

void CSurface3D::SetVertexRef(unsigned p_index, int p_vertexRefID, float p_s, float p_t)
{
	vertexRefs[p_index] = p_vertexRefID;
	texCoords[p_index].s = p_s;
	texCoords[p_index].t = p_t;
}

void CSurface3D::SetVertexTexCoords(unsigned p_index, float p_s, float p_t)
{
	texCoords[p_index].s = p_s;
	texCoords[p_index].t = p_t;
}

void CSurface3D::SetVertexLMCoords(unsigned p_index, float p_sLM, float p_tLM)
{
	texCoordsLM[p_index].s = p_sLM;
	texCoordsLM[p_index].t = p_tLM;
}

void CSurface3D::SetTexture(char *p_name)
{
	texture = APITextures.GetTexture(p_name);
}

bool CSurface3D::SphereCollision(CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CModelVertex3D *p_vertices, CVector3D &p_collisionNormalOut)
{
	float collisionTime;
	float tmp;
	CVector3D tmpPos;
	CVector3D tmpV1;
	CVector3D tmpV2;
	CVector3D tmpCP;
	unsigned i;

	// Make sure sphere is moving in a direction that would make it collide
	tmp = p_offset * normal;
	// Moving away? (tolerance for up-close collisions)
//	if (tmp >= -0.0001)
	if (tmp > 0)
		return false;
	// Moving parallel?
	if (fabs(tmp) < 0.00001f)
		return false;

	collisionTime = (p_radius - ((p_pos - p_vertices[vertexRefs[0]].vertex) * normal)) / tmp;

	// this -p_radius/1000 is to keep sized objects from falling through floors
	// we shouldn't need it for points or segments, I hope...
	if ((collisionTime > (-p_radius / 1000.0f)) && (collisionTime <= 1.0f))
	{
		if (collisionTime < p_timeOut)
		{
			// Still need to make sure the collision is inside the surface!
			//   If it is, no need to check against segments or points,
			//   it collides witht he body of the polygon first
			tmpPos = p_pos + (p_offset * collisionTime);
			tmpPos = tmpPos - (normal * p_radius);
			// tmpPos is now the point of collision
        
			// Now make sure tmpPos is actually inside the surface
			// NOTE: This assumes the surface is convex and CW!!!
			if (PointInside(tmpPos, p_vertices))
			{
				p_timeOut = collisionTime;
				p_collisionNormalOut = normal;
				//Debug.Print "Surface collision"
				return true;
			}
		}
	}

	// If a collision with the polygon itself was successful, then
	//   there is no need to check the segments, since it would
	//   collide with the surface first
	// However, if we're here, then we know the collision with the surface failed
	// To check for a collision with the segments, the unit vector for
	//   the segment will be used to transform the check to one plane,
	//   then a Sphere-point check will be performed

	// ?? Note: this is checking all the segments of each surface,
	//   so there WILL be repeat checks.  In the future, I may want
	//   to assemble a segment list for a model

	CVector3D newPos;
	CVector3D newOffset;
	CVector3D tmpCollisionNormal;
	float tmpTime;

	float o1;

	// Find the segment we collide with first time-wise
	// ?? Couldn't we just find the first success?  Then we would return
	//    inside the loop
	bool result = false;
	for (i=0; i<vertexRefQty; i++)
	{
		newPos = (p_pos + (segUnits[i] * ((p_vertices[vertexRefs[i]].vertex - p_pos) * (segUnits[i]))));
		newOffset = (p_offset - (segUnits[i] * (p_offset * segUnits[i])));
    
		tmpTime = p_timeOut;
    
		if (p_vertices[vertexRefs[i]].vertex.SphereCollision(newPos, newOffset, p_radius, tmpTime, tmpCollisionNormal))
		{
			// tmpCollisionNormal is actually correct, so is tmpTime
			// But we need to be sure that the Sphere actually
			//   collides within the segment...
        
			// Re-use temps
			newPos = (p_pos + (p_offset *tmpTime));
        
			o1 = (newPos - p_vertices[vertexRefs[i]].vertex) * segUnits[i];
			if ((o1 >= 0) && (o1 <= segLengths[i]))
			{
				// We have a valid collision, whew
				result = true;
				p_timeOut = tmpTime;
				p_collisionNormalOut = tmpCollisionNormal;
			}
		}
	}

	return result;
}

// Note: for this to work, the surface must have had CalculateSurfaceData() called
//   on it
bool CSurface3D::PointInside(CVector3D &p_point, CModelVertex3D *p_vertices)
{
	unsigned i;

    // Note: This assumes the surface is convex!
    //   Also, there may be a quicker way to do this
    // ??
    for (i=0; i<vertexRefQty; i++)
	{
		// Watch for the point to be outside a segment.  If we find one,
		//    the point is not inside
        if (((p_vertices[vertexRefs[i]].vertex - p_point) * segNormals[i]) > 0.0f)
			return false;
	}
    
    return true;
}

// If you want collision distance to be accurate, make sure p_directionUnit is length 1
bool CSurface3D::RayCollision(CVector3D &p_start, CVector3D &p_directionUnit, float &p_collisionDistanceOut, CModelVertex3D *p_vertices, CVector3D &p_collisionPointOut)
{
	float tmp;
	float k;
	CVector3D tmpPos;

	// Make sure ray is moving in a direction that would make it collide
	tmp = p_directionUnit * normal;
	if (tmp >= 0.0f) return false;
	if (fabs(tmp) < 0.0000001f) return false;

	k = -(normal * (p_start - (p_vertices[vertexRefs[0]].vertex))) / (normal * p_directionUnit);

	if (k < 0.0f) return false; // Collision is behind starting point

	if (k < p_collisionDistanceOut)
	{
		tmpPos = p_start + (p_directionUnit * k);

		// Make sure point is inside surface
		if (PointInside(tmpPos, p_vertices))
		{
			p_collisionDistanceOut = k;
			p_collisionPointOut = tmpPos;
			return true;
		}
	}

	return false;
}

// NOTE: p_decalSurface should have CalculateSurfaceData() called before
//   entering this function, so that SurfaceIntersection's call of
//   PointInside provides an accurate result
void CSurface3D::PlaceDecal(CSurface3D &p_decalSurface, CModelVertex3D *p_decalVertices, CModelVertex3D *p_verticesThis, CSurfaceDecalList *p_decalLists, CSurfaceDecalList *p_decalLists2)
{
	CVector3D vertOut[32];
	unsigned vertexQty = 0;
	unsigned i;
	float s, t;

	if (decalListID == -1) return;
	if (p_decalLists == NULL) return;

	if (!SurfaceIntersection(p_decalSurface, p_decalVertices, p_verticesThis, vertOut, vertexQty)) return;

	if (vertexQty > 0)
	{
		// Add the decal
		CSurfaceDecalListNode *patchListNode;
		patchListNode = p_decalLists[decalListID].NewNode();
		patchListNode->Initialize (vertexQty);
		for (i=0; i<vertexQty; i++)
			patchListNode->GetVertex(i).vertex.SetVec(vertOut[i].x, vertOut[i].y, vertOut[i].z);
    
		if (p_decalSurface.GetTexture())
		{
			patchListNode->GetSurface().Initialize(vertexQty, p_decalSurface.GetTexture());
			for (i=0; i<vertexQty; i++)
			{
				p_decalSurface.CalculateTex1Coords(patchListNode->GetVertex(i).vertex, s, t);
				patchListNode->GetSurface().SetVertexRef(i, i, s, t);
			}
		}
		else
		{
			patchListNode->GetSurface().Initialize(vertexQty, p_decalSurface.GetTexture());
			for (i=0; i<vertexQty; i++)
				patchListNode->GetSurface().SetVertexRef(i, i);
		}
    
		patchListNode->GetSurface().SetAlpha(p_decalSurface.GetAlpha());
    
		if (p_decalSurface.UseSurfaceColor())
			patchListNode->GetSurface().SetColor(p_decalSurface.color);
    
		if (lightmapID != -1)
		{
			patchListNode->GetSurface().SetLightmapID(lightmapID);
			for (i=0; i<vertexQty; i++)
			{
				CalculateTexLMCoords(patchListNode->GetVertex(i).vertex, s, t);
				patchListNode->GetSurface().SetVertexLMCoords(i, s, t);
			}
			patchListNode->GetSurface().SetDynLMOrigin(dynLMX, dynLMY);
		}
    
		patchListNode->backside = false;
    
		// Now, if this surface has an opposing surface,
		//   we need to add this decal to that one too.
		// It will only get rendered however if this surface
		//   is not a mirror and we are in front
		int listID;
		if (p_decalLists2)
		if (opposingSurface)
		if ((listID=opposingSurface->GetDecalListID()) != -1)
		{
			patchListNode = p_decalLists2[listID].NewNode();
        
			patchListNode->Initialize (vertexQty);
			for (i=0; i<vertexQty; i++)
				patchListNode->GetVertex(i).vertex.SetVec(vertOut[i].x, vertOut[i].y, vertOut[i].z);
        
			if (p_decalSurface.GetTexture())
			{
				patchListNode->GetSurface().Initialize(vertexQty, p_decalSurface.GetTexture());
				for (i=0; i<vertexQty; i++)
				{
					p_decalSurface.CalculateTex1Coords(patchListNode->GetVertex(i).vertex, s, t);
					patchListNode->GetSurface().SetVertexRef(i, i, s, t);
				}
			}
			else
			{
				patchListNode->GetSurface().Initialize(vertexQty, p_decalSurface.GetTexture());
				for (i=0; i<vertexQty; i++)
					patchListNode->GetSurface().SetVertexRef(i, i);
			}
        
			patchListNode->GetSurface().SetAlpha(p_decalSurface.GetAlpha());
        
			if (p_decalSurface.UseSurfaceColor())
				patchListNode->GetSurface().SetColor(p_decalSurface.color);
        
			if (opposingSurface->GetLightmapID() != -1)
			{
				patchListNode->GetSurface().SetLightmapID(opposingSurface->GetLightmapID());
				for (i=0; i<vertexQty; i++)
				{
					opposingSurface->CalculateTexLMCoords(patchListNode->GetVertex(i).vertex, s, t);
					patchListNode->GetSurface().SetVertexLMCoords(i, s, t);
				}
				patchListNode->GetSurface().SetDynLMOrigin(dynLMX, dynLMY);
			}
        
			patchListNode->backside = true;
		} // if opposingSurface
    
	} // if vertexQty > 0

}

void CSurface3D::CalculateTex1Coords(CVector3D &p_point, float &p_sOut, float &p_tOut)
{
    CVector3D v1;
    v1 = p_point - Tex1Origin;
    p_sOut = v1 * Tex1AxisS;
    p_tOut = v1 * Tex1AxisT;
}

void CSurface3D::CalculateTexLMCoords(CVector3D &p_point, float &p_sOut, float &p_tOut)
{
    CVector3D v1;
    v1 = p_point - LMOrigin;
    p_sOut = v1 * LMAxisS;
    p_tOut = v1 * LMAxisT;
}

void CSurface3D::CalculateTex1Axes(CModelVertex3D *p_vertices)
{
	// Uses points 0, 1 and 2 to calculate the origin and axes
	//   needed for calculating a new texture coordinate from an
	//   arbitrary surface point (actually we are now looking
	//   for the first two texture segments that aren't on the same
	//   line)

	// We end up with a system of equations solved by using
	//   an inverse matrix.  Check your linear algebra books!
	//   We solve for (-s0, -t0) to move point 0 to its origin
	//  then solve for (1-s0, -t0) and (-s0, 1-t0) to get the
	//   texture axes

	// Inverse of | a b | is | d/k -b/k | where k = ad-bc
	//            | c d |    | -c/k a/k |

	// Then the inverse matrix * what we are solving for =
	//    the scalars we need to multiply v0 and v1 by to get
	//    to our destination!

	// Find the first sequential pair of segments not on the same line

	float a, b, c, d;
	float k, k1, k2;
	CVector3D v0;
	CVector3D v1;

	unsigned int i;
	for (i=0; i<=vertexRefQty-3; i++)
	{
		a = texCoords[i+1].s - texCoords[i+0].s;
		b = texCoords[i+2].s - texCoords[i+1].s;
		c = texCoords[i+1].t - texCoords[i+0].t;
		d = texCoords[i+2].t - texCoords[i+1].t;
		k = a * d - b * c;

		// if this fails, the segments are on the same line
		if (fabs(k) > 0.00001f)
		{
			v0 = p_vertices[vertexRefs[i+1]].vertex - p_vertices[vertexRefs[i+0]].vertex;
			v1 = p_vertices[vertexRefs[i+2]].vertex - p_vertices[vertexRefs[i+1]].vertex;
    
			// Solving for (-s0, -t0) for origin
			k1 = (d / k) * (-texCoords[i+0].s) + (-b / k) * (-texCoords[i+0].t);
			k2 = (-c / k) * (-texCoords[i+0].s) + (a / k) * (-texCoords[i+0].t);
			// Setting origin
			Tex1Origin = p_vertices[vertexRefs[i+0]].vertex + (v0 * k1) + (v1 * k2);
    
			// Solving for (1-s0, -t0) for axis along s
			k1 = (d / k) * (1 - texCoords[i+0].s) + (-b / k) * (-texCoords[i+0].t);
			k2 = (-c / k) * (1 - texCoords[i+0].s) + (a / k) * (-texCoords[i+0].t);
			// Setting S Axis
			Tex1AxisS = p_vertices[vertexRefs[i+0]].vertex + (v0 * k1) + (v1 * k2);
			Tex1AxisS = Tex1AxisS - Tex1Origin;
			Tex1AxisS = Tex1AxisS * (1.0f / Tex1AxisS.MgnTudSqr());
    
			// Solving for (-s0, 1-t0) for axis along t
			k1 = (d / k) * (-texCoords[i+0].s) + (-b / k) * (1 - texCoords[i+0].t);
			k2 = (-c / k) * (-texCoords[i+0].s) + (a / k) * (1 - texCoords[i+0].t);
			// Setting T Axis
			Tex1AxisT = p_vertices[vertexRefs[i+0]].vertex + (v0 * k1) + (v1 * k2);
			Tex1AxisT = Tex1AxisT - Tex1Origin;
			Tex1AxisT = Tex1AxisT * (1.0f / Tex1AxisT.MgnTudSqr());

			// Whew, done
			return;
		}
	}

}

// Get segment intersection between segment and a segment of this surface
// There should be only one valid collision to find
//    from inside or outside
// if p_outside is false, collision must occur from the inside, else it must occur from the outside
//   (This doesn't meant he start point is on the inside or outside, it just
//    means we are looking for the collision to occur on the outside or inside surface)
// Segment to collide defined by p_start, p_directionUnit and p_segLength
bool CSurface3D::GetSegmentIntersection(CVector3D &p_start, CVector3D &p_directionUnit, float p_segLength, bool p_outside, CModelVertex3D *p_vertices, CVector3D &p_intersectionPointOut, unsigned &p_segIDOut)
{
	unsigned i;
	CVector3D v1;
	CVector3D segStart;
	float dp, k;

	bool ok;
	for (i=0; i<vertexRefQty; i++)
	{
		ok = true;
		if (p_outside)
		{
			if (((p_start - p_vertices[vertexRefs[i]].vertex) * segNormals[i]) > 0)
				ok = false;
		}
		else
		{
			if (((p_start - p_vertices[vertexRefs[i]].vertex) * segNormals[i]) < 0)
				ok = false;
		}

		if (ok)
		{
			segStart = p_vertices[vertexRefs[i]].vertex;
			dp = segNormals[i] * p_directionUnit;
			if (fabs(dp) > 0.00001f)
			{
				k = -(segNormals[i] * (p_start - segStart)) / dp;
				if ((k >= 0) && (k <= p_segLength))
				{
					// Make sure collision point is in the segment
					v1 = p_start + (p_directionUnit * k);
					dp = (v1 - segStart) * segUnits[i];
					if ((dp >= 0) && (dp <= segLengths[i]))
					{
						// Correct v1, force it on to the surface
						p_intersectionPointOut = v1 - (normal * ((v1 - segStart) * normal));
						p_segIDOut = i;

						return true;
					}
				}
			}
		}
	}

	return false;
}

// Intersect two surfaces
// Return an array of points making up the result, p_vertexQtyOut = # vertices
// This surface has priority - if there is any variance in the
//    height of the vertices relative to this surface, the output vertices
//    will be on this surface
// It is also assumed that both surfaces are basically defined
//    clockwise near to having the same normal
// Both surfaces should have called both their
//    CalculateSurfaceNormal routines so that all their
//    segment values are set up
// So far, this routine is only used to prepare a patch
//    for rendering on another surface
// Returns false if number of output vertices is 0

// Note: SurfaceP should have CalculateSurfaceData() called on it before
//    entering this function, so that PointInside at the end provides an
//    accurate result
bool CSurface3D::SurfaceIntersection(CSurface3D &p_surfaceP, CModelVertex3D *p_verticesP, CModelVertex3D *p_verticesThis, CVector3D *p_vertOut, unsigned &p_vertexQtyOut)
{
	bool exitPoint;
	bool entryPoint;
	bool lastPointInside;
	bool firstPointInside;
	unsigned storedEntrySeg;
	unsigned exitSeg, entrySeg;
	unsigned currentSeg, targetSeg;

	CVector3D point;
	CVector3D start;
	float length;
	CVector3D direction;

/*	 The proper algorithm:
	 For each point in the surface
	    if point is inside
	       if first point then
	           firstpointinside = true
	       else
	           if there's an exit point
	               get entry point and seg (seg = last point to current)
	               trace points from exit seg to entry seg
	               add entry point to list
	               exit point = false
	           else
	'               if last point wasn't inside
	'                   get entry point and seg
	'                   add entry point to list
	'                   store entry point and seg (will be handled with final point)
	'       add this point to the list
	'       lastpointinside = true
	'       if this is the final point
	'           if first point was not inside
	'               get exit point (seg = current to 0)
	'               add exit point to list
	'               trace points from exit seg to stored entry seg
	'    else
	'       if first point then
	'           firstpointinside = false
	'       else
	'           if last point was inside
	'               get exit point and seg (seg = last to current)
	'               add exit point to list
	'               store exit seg
	'               flag exit point
	'           else
	'               check for entry point (seg = last to current)
	'               if one found
	'                   if there's an exit point then
	'                       trace points from exit seg to entry seg
	'                   else
	'                       store entry seg
	'                   add entry point to list
	'                   get exit point (if not found, invalid list)
	'                   add exit point to list
	'                   store exit seg
	'                   flag exit point
	'       lastpointinside = false
	'       if this is the final point
	'           if first point was inside then
	'               if no exit point, list invalid
	'               get entry point (seg = current to 0)
	'               trace points from exit seg to entry seg
	'               add entry point
	'           else
	'               check for entry point (seg=current to 0)
	'               if found
	'                   if exit point
	'                       trace from exit to entry
	'                   add entry point
	'                   if no stored entry
	'                       store this now
	'                   get exit point
	'                   add exit point
	'                   store exit seg
	'                   flag exit point
	'               if there is a stored exit and entry point
	'                   trace from exit to stored entry segs
	'
	' After loop: if no points in list but a point from
	'    the main surface is in the passed surface,
	'    create list from main surface and return that
*/

	p_vertexQtyOut = 0;
	exitPoint = false;
	entryPoint = false;
	lastPointInside = false;
	firstPointInside = false;
	storedEntrySeg = -1;

	unsigned i;
	for (i=0; i<p_surfaceP.GetVertexRefQty(); i++)
	{
		if (PointInside(p_verticesP[p_surfaceP.GetVertexRef(i)].vertex, p_verticesThis))
		{
			if (i == 0)
				firstPointInside = true;
			else
			{
				if (exitPoint)
				{
					// If we have an exit point from before,
					//    get the entry point and add vertices around the
					//    this surface until we get to the entry point
					start = p_verticesP[p_surfaceP.GetVertexRef(i - 1)].vertex;
					direction = p_verticesP[p_surfaceP.GetVertexRef(i)].vertex - start;
					length = direction.MgnTud();
					direction.Unitize();
					if (GetSegmentIntersection(start, direction, length, true, p_verticesThis, point, entrySeg))
					{
						// Trace points from exitSeg to entrySeg
						currentSeg = exitSeg;
						targetSeg = entrySeg;
						while (currentSeg != targetSeg)
						{
							currentSeg = currentSeg + 1;
							if (currentSeg >= vertexRefQty)
								currentSeg = 0;

							// Add point(currentSeg) to list
							p_vertOut[p_vertexQtyOut] = p_verticesThis[vertexRefs[currentSeg]].vertex;
							p_vertexQtyOut = p_vertexQtyOut + 1;
						}
                    
						// Add entry point to list
						p_vertOut[p_vertexQtyOut] = point;
						p_vertexQtyOut = p_vertexQtyOut + 1;
					} // got an intersection with a segment
					else
					{
						// No entry point, list invalid
						p_vertexQtyOut = 0;
						return false;
					}
                
					exitPoint = false;
				}	// exitPoint
				else
				{
					if (!lastPointInside)
					{
						// We are coming in from the outside,
						//   but we don't have an exit point yet
						// Get entry point, add that to the list,
						//   Store entrySeg.  We'll take care
						//   of it at the end of the loop
						start = p_verticesP[p_surfaceP.GetVertexRef(i - 1)].vertex;
						direction = p_verticesP[p_surfaceP.GetVertexRef(i)].vertex - start;
						length = direction.MgnTud();
						direction.Unitize();
						if (GetSegmentIntersection(start, direction, length, true, p_verticesThis, point, storedEntrySeg))
						{
							// Add point(currentSeg) to list
							p_vertOut[p_vertexQtyOut] = point;
							p_vertexQtyOut = p_vertexQtyOut + 1;
						}
						else
						{
							// No entry point, list invalid
							p_vertexQtyOut = 0;
							return false;
						}
					}
				} // no exit point
			} // not first point
    
			// Add vertex to output
			p_vertOut[p_vertexQtyOut] = p_verticesP[p_surfaceP.GetVertexRef(i)].vertex;
			p_vertexQtyOut = p_vertexQtyOut + 1;
			lastPointInside = true;
        
			if (i == (p_surfaceP.GetVertexRefQty() - 1))
			{
				if (!firstPointInside)
				{
					// get exit point
					// add exit point to list
					start = p_verticesP[p_surfaceP.GetVertexRef(p_surfaceP.GetVertexRefQty() - 1)].vertex;
					direction = p_verticesP[p_surfaceP.GetVertexRef(0)].vertex - start;
					length = direction.MgnTud();
					direction.Unitize();
					if (GetSegmentIntersection(start, direction, length, false, p_verticesThis, point, exitSeg))
					{
						// Add point to list
						p_vertOut[p_vertexQtyOut] = point;
						p_vertexQtyOut = p_vertexQtyOut + 1;
					}
					else
					{
						// No exit point, list invalid
						p_vertexQtyOut = 0;
						return false;
					}
                
					// if no storedEntrySeg then list invalid
					if (storedEntrySeg == -1)
					{
						p_vertexQtyOut = 0;
						return false;
					}
                
					// trace points from exit seg to stored entry seg
					currentSeg = exitSeg;
					targetSeg = storedEntrySeg;
					while (currentSeg != targetSeg)
					{
						currentSeg = currentSeg + 1;
						if (currentSeg >= vertexRefQty)
							currentSeg = 0;

						// Add point(currentSeg) to list
						p_vertOut[p_vertexQtyOut] = p_verticesThis[vertexRefs[currentSeg]].vertex;
						p_vertexQtyOut = p_vertexQtyOut + 1;
					}
				}	// first point was inside
			} // on last point
		}	// point inside
		else
		{
			if (i == 0)
			{
				firstPointInside = false;
			}
			else
			{
				if (lastPointInside)
				{
					// Get exit point and add that to the list
					// Store exit seg #
					start = p_verticesP[p_surfaceP.GetVertexRef(i - 1)].vertex;
					direction = p_verticesP[p_surfaceP.GetVertexRef(i)].vertex - start;
					length = direction.MgnTud();
					direction.Unitize();
					if (GetSegmentIntersection(start, direction, length, false, p_verticesThis, point, exitSeg))
					{
						// Add exit point to list
						p_vertOut[p_vertexQtyOut] = point;
						p_vertexQtyOut = p_vertexQtyOut + 1;
					}
					else
					{
						// No exit point, list invalid
						p_vertexQtyOut = 0;
						return false;
					}
                
					exitPoint = true;
				}
				else
				{
					// Check for seg intersection (entry).  If so,
					//   we have an entry point - trace with previous
					//   exitpoint if there is one, otherwise set
					//   entrypoint
					start = p_verticesP[p_surfaceP.GetVertexRef(i - 1)].vertex;
					direction = p_verticesP[p_surfaceP.GetVertexRef(i)].vertex - start;
					length = direction.MgnTud();
					direction.Unitize();
					if (GetSegmentIntersection(start, direction, length, true, p_verticesThis, point, entrySeg))
					{
						if (exitPoint)
						{
							// Trace them
							currentSeg = exitSeg;
							targetSeg = entrySeg;
							while (currentSeg != targetSeg)
							{
								currentSeg = currentSeg + 1;
								if (currentSeg >= vertexRefQty)
									currentSeg = 0;

								// Add point(currentSeg) to list
								p_vertOut[p_vertexQtyOut] = p_verticesThis[vertexRefs[currentSeg]].vertex;
								p_vertexQtyOut = p_vertexQtyOut + 1;
							}
						}
						else
							storedEntrySeg = entrySeg;
                    
						// Add entry point to list
						p_vertOut[p_vertexQtyOut] = point;
						p_vertexQtyOut = p_vertexQtyOut + 1;
                    
						// And then, we have an exit point (we better)
						start = p_verticesP[p_surfaceP.GetVertexRef(i - 1)].vertex;
						direction = p_verticesP[p_surfaceP.GetVertexRef(i)].vertex - start;
						length = direction.MgnTud();
						direction.Unitize();
						if (GetSegmentIntersection(start, direction, length, false, p_verticesThis, point, exitSeg))
						{
							// Add exit point to output
							p_vertOut[p_vertexQtyOut] = point;
							p_vertexQtyOut = p_vertexQtyOut + 1;
						}
						else
						{
							p_vertexQtyOut = 0;
							return false;
						}
                    
						exitPoint = true;
					}
				}	// last point not inside
			}	// not first point
        
			lastPointInside = false;
        
			if (i == (p_surfaceP.GetVertexRefQty() - 1))
			{
				if (firstPointInside)
				{
					// if no exitpoint, list invalid
					if (!exitPoint)
					{
						p_vertexQtyOut = 0;
						return false;
					}
					else
					{
						// get entry point (seg = current to 0)
						start = p_verticesP[p_surfaceP.GetVertexRef(p_surfaceP.GetVertexRefQty() - 1)].vertex;
						direction = p_verticesP[p_surfaceP.GetVertexRef(0)].vertex - start;
						length = direction.MgnTud();
						direction.Unitize();
						if (GetSegmentIntersection(start, direction, length, true, p_verticesThis, point, entrySeg))
						{
							// trace points from exit seg to entry seg
							currentSeg = exitSeg;
							targetSeg = entrySeg;
							while (currentSeg != targetSeg)
							{
								currentSeg = currentSeg + 1;
								if (currentSeg >= vertexRefQty)
									currentSeg = 0;

								// Add point(currentSeg) to list
								p_vertOut[p_vertexQtyOut] = p_verticesThis[vertexRefs[currentSeg]].vertex;
								p_vertexQtyOut = p_vertexQtyOut + 1;
							}
                        
							// add entry point
							p_vertOut[p_vertexQtyOut] = point;
							p_vertexQtyOut = p_vertexQtyOut + 1;
						}
						else
						{
							// no entry, list invalid
							p_vertexQtyOut = 0;
							return false;
						}
                    
						// We're done
                    
					}
				}
				else
				{
					// check for entry point
					start = p_verticesP[p_surfaceP.GetVertexRef(p_surfaceP.GetVertexRefQty() - 1)].vertex;
					direction = p_verticesP[p_surfaceP.GetVertexRef(0)].vertex - start;
					length = direction.MgnTud();
					direction.Unitize();
					if (GetSegmentIntersection(start, direction, length, true, p_verticesThis, point, entrySeg))
					{
						// if found
						// if there's an exit point
						if (exitPoint)
						{
							// trace from exit to entry
							currentSeg = exitSeg;
							targetSeg = entrySeg;
							while (currentSeg != targetSeg)
							{
								currentSeg = currentSeg + 1;
								if (currentSeg >= vertexRefQty)
									currentSeg = 0;

								// Add point(currentSeg) to list
								p_vertOut[p_vertexQtyOut] = p_verticesThis[vertexRefs[currentSeg]].vertex;
								p_vertexQtyOut = p_vertexQtyOut + 1;
							}
						}
                    
						// add entry point
						p_vertOut[p_vertexQtyOut] = point;
						p_vertexQtyOut = p_vertexQtyOut + 1;
                    
						// if no stored entry
						// store this now
						if (storedEntrySeg == -1)
							storedEntrySeg = entrySeg;
                    
						// get exit point
						start = p_verticesP[p_surfaceP.GetVertexRef(p_surfaceP.GetVertexRefQty() - 1)].vertex;
						direction = p_verticesP[p_surfaceP.GetVertexRef(0)].vertex - start;
						length = direction.MgnTud();
						direction.Unitize();
						if (GetSegmentIntersection(start, direction, length, false, p_verticesThis, point, exitSeg))
						{
							// add exit point
							p_vertOut[p_vertexQtyOut] = point;
							p_vertexQtyOut = p_vertexQtyOut + 1;
							exitPoint = true;
						}
						else
						{
							// List invalid, we should have had one
							p_vertexQtyOut = 0;
							return false;
						}
					}
                    
					// if there is a stored exit and entry point
						// trace from exit to stored entry segs
					if ((storedEntrySeg != -1) && (exitPoint))
					{
						currentSeg = exitSeg;
						targetSeg = storedEntrySeg;
						while (currentSeg != targetSeg)
						{
							currentSeg = currentSeg + 1;
							if (currentSeg >= vertexRefQty)
								currentSeg = 0;

							// Add point(currentSeg) to list
							p_vertOut[p_vertexQtyOut] = p_verticesThis[vertexRefs[currentSeg]].vertex;
							p_vertexQtyOut = p_vertexQtyOut + 1;
						}
					}
				}
			}	// on last point
		} // point outside
	}	// for i

	// If there are no output points, we need to see if any of this
	//   surface's points are in the P surface (we only need to check one)
	//   If so then P totally encompasses this surface, so send
	//   all this surface's points back
	if (p_vertexQtyOut == 0)
	{
		if (p_surfaceP.PointInside(p_verticesThis[vertexRefs[0]].vertex, p_verticesP))
		{
			for (i=0; i<vertexRefQty; i++)
			{
				p_vertOut[p_vertexQtyOut] = p_verticesThis[vertexRefs[i]].vertex;
				p_vertexQtyOut = p_vertexQtyOut + 1;
			}
		}
	}

	// Send back vertices
	if (p_vertexQtyOut == 0)
		return false;
	else
		return true;

}

//////////////////////////////////////////////////////
// Lightmap functions

void CSurface3D::ClearLightmap(CColorRGB p_color, CAPITexture **p_lightmaps)
{
    if ((lightmapID != -1) && (p_lightmaps))
        p_lightmaps[lightmapID]->ClearLightmap(p_color);
}

void CSurface3D::ApplyRadiusLightLM(float p_x, float p_y, float p_radius, CColorRGB p_color, CAPITexture **p_lightmaps)
{   
	if ((lightmapID != -1) && (p_lightmaps))
        p_lightmaps[lightmapID]->ApplyRadiusLight(p_x, p_y, p_radius, p_color);
}

void CSurface3D::ApplyLightSourceLM(CLightSource *p_ls, CModelVertex3D *p_vertices, CAPITexture **p_lightmaps)
{   
	if ((lightmapID != -1) && (p_lightmaps))
	if ((p_ls->visible) & (p_ls->intensity > 0))
	{
		switch (p_ls->type)
		{
		case APILIGHTTYPE_SPHERICALLIGHTMAP:
			float distance = (p_ls->position - p_vertices[vertexRefs[0]].vertex) * normal;
			if ((distance > -(p_ls->radius / 8.0f)) && ( distance < p_ls->radius))
			{
				if (distance < 0)
					distance = distance * p_ls->radius / -(p_ls->radius/8.0f);
				float radius = float(sqrt(pow(p_ls->radius,2) - pow(distance,2)));
				CColorRGB color;
				color = p_ls->color;
				float x, y;
				CalculateTexLMCoords(p_ls->position, x, y);
				p_lightmaps[lightmapID]->ApplyRadiusLight(x, y, radius, distance, color);
			}
		break;
		}
	}
}

void CSurface3D::SetupLightmap(CModelVertex3D *p_vertices, CAPITexture **p_lightmaps, float p_texelPerUnit, CPixelStorageMgr *p_pixelStorageMgr, unsigned *p_dynLMID)
{
    unsigned p_widthOut, p_heightOut;
    
    if (lightmapID == -1) return;
	if (!texCoordsLM) return;
	if (!p_vertices) return;
    
    // Calculate area for lightmap
    CalculateLightmapArea(p_vertices, p_texelPerUnit, p_widthOut, p_heightOut, p_pixelStorageMgr, p_dynLMID);
    p_lightmaps[lightmapID]->SetLightmap(true);
    p_lightmaps[lightmapID]->SetMipmap(false);
    
    // Initialize the lightmap texture storage and give
    //   it an ID for the API
    p_lightmaps[lightmapID]->SetLightmapDimensions(p_widthOut, p_heightOut, p_texelPerUnit);
    
    // Set up the surface's texture coordinates for the lightmap
    unsigned i;
    float s, t;
    for (i=0; i<vertexRefQty; i++)
	{
        CalculateTexLMCoords(p_vertices[vertexRefs[i]].vertex, s, t);
        if (s < 0) s = 0;
        if (t < 0) t = 0;
        if (s > 1) s = 1;
        if (t > 1) t = 1;
        texCoordsLM[i].s = s;
        texCoordsLM[i].t = t;
	}
}

void CSurface3D::CalculateLightmapArea(CModelVertex3D *p_vertices, float p_texelPerUnit, unsigned &p_widthOut, unsigned &p_heightOut, CPixelStorageMgr *p_pixelStorageMgr, unsigned *p_dynLMID)
{
	// Loop through all possibilities
	// Get first side, use that as the primary axis
	// Then use the next side to calculate an orthogonal axis
	// Unitize them and use them to calculate the bounds of
	//   the box holding all the points of the polygon
	// Actually, we need a box with the minimum values of the
	//   next powers of 2
	// Find the combination of side and next side that
	//   results in the smallest needed lightmap

	CVector3D v1, v2;
	CVector3D v3;
	CVector3D axis1, axis2;
	float k1min, k1max;
	float k2min, k2max;
	float area;
	CVector3D bestAxis1, bestAxis2;
	unsigned i, j;
	float xd, yd;
	float bestXd, bestYd;
	CVector3D bestLMOrigin;

	float cp;

	v2 = p_vertices[vertexRefs[0]].vertex - p_vertices[vertexRefs[vertexRefQty - 1]].vertex;
	area = -1.0f;
	for (i=0; i<vertexRefQty; i++)
	{
		v1 = v2;
		if (i < vertexRefQty - 1)
			v2 = p_vertices[vertexRefs[i + 1]].vertex - p_vertices[vertexRefs[i]].vertex;
		else
			v2 = p_vertices[vertexRefs[0]].vertex - p_vertices[vertexRefs[vertexRefQty - 1]].vertex;
    
		// This combo ensures that we get two adjacent segments
		//   that can make a set of axes, so parallel adjacent
		//   axes will be ignored
		axis1 = v1;
		if (axis1.Unitize())
		{
			cp = (v2.CrsProd(axis1)).MgnTud();
			if (cp > 0.0001f)
			{
				axis2 = v2 - (axis1 * (v2 * axis1));
				if (axis2.Unitize())
				{
					// axis1 and axis2 are now unit orthogonal axes for the
					//   polygon.  Get the bounding box.
					k1min = 1000000;
					k2min = 1000000;
					k1max = -1000000;
					k2max = -1000000;
					for (j=0; j<vertexRefQty; j++)
					{
						v3 = p_vertices[vertexRefs[j]].vertex - p_vertices[vertexRefs[i]].vertex;
						if ((v3 * axis1) < k1min) k1min = v3 * axis1;
						if ((v3 * axis1) > k1max) k1max = v3 * axis1;
						if ((v3 * axis2) < k2min) k2min = v3 * axis2;
						if ((v3 * axis2) > k2max) k2max = v3 * axis2;
					}
    
					// We have our bounding box
    
					// This adjustment is for the common problem of blending
					//    clamped textures - it avoids the dark borders around
					//    the edges by hiding the edge texels beyond the surface
					//    bounds.  Clever, huh?
					// All this does is make sure the size we need is 2 more
					//    on each axis.  It has no other effect.
					// However, this also means the LMOrigin point will have
					//    to be adjusted by 1 texel as well.  See below
					k1max = k1max + 2;
					k2max = k2max + 2;
    
					// Now calculate the size of the texture we'd need to
					//   make a lightmap (dimensions must be powers of 2, and at least 4
					//   so that OpenGL doesn't gack)
					// Note that -int(-number) = its ceiling in VB
					// In C++ we can just use ceil()
					xd = float(pow(2.0f, ceil(log((k1max - k1min) * p_texelPerUnit) / log(2.0f))));
					yd = float(pow(2.0f, ceil(log((k2max - k2min) * p_texelPerUnit) / log(2.0f))));
					if (xd < 4) xd = 4;
					if (yd < 4) yd = 4;
					// Don't let the end texture be larger than the dynamic lightmap.
					if (xd > APITextures.GetDynamicLightmap()->GetWidthf()) xd = APITextures.GetDynamicLightmap()->GetWidthf();
					if (yd > APITextures.GetDynamicLightmap()->GetHeightf()) yd = APITextures.GetDynamicLightmap()->GetHeightf();
					// Note that any adjustment of xd or yd will throw off
					//    the accuracy of axis1 and axis2 (I think)
    
					if ((area < 0) || ((xd * yd) < area))
					{
						bestXd = xd;
						bestYd = yd;
						area = bestXd * bestYd;
						bestAxis1 = axis1;
						bestAxis2 = axis2;
						bestLMOrigin = p_vertices[vertexRefs[i]].vertex + (axis1 * k1min);
						// Adjust LMOrigin by 1 texel, too - see above
						bestLMOrigin = bestLMOrigin - (axis1 * (1.0f / p_texelPerUnit)) - (axis2 * (1.0f / p_texelPerUnit));
						//bestk1min = k1min
						//bestk2min = k2min
					}
    
					// NOTE: The world space lighting routine will need to know the
					//   location and orientation of the lightmap in world space
					//   in order to light it (that's what CalculateLMTexCoords does with
					//   the point on the surface)
    
					// ?? Also, this routine doesn't make sure the lightmap
					//   is vertical (like Quake 2 maps are set up)

				}	// Unitize 2 ok
			}
    
		}	// Unitize 1 ok
	}	// for i

	//MsgBox area & ", " & bestXd & ", " & bestYd
	// Set outputs and texture origin and scale vectors
	p_widthOut = unsigned(bestXd + 0.01f);
	p_heightOut = unsigned(bestYd + 0.01f);

	// The way these are set up, take any point on the surface,
	//   subtract LMOrigin, then dotproduct that with
	//   LMAxisS for the lightmap's s, and LMAxisT for the lightmap's t (0.0-1.0)
	//   Multiply the resulting s and t by width and height to get pixel values
	LMOrigin = bestLMOrigin;
	LMAxisS = bestAxis1 * (1.0f / (bestXd / p_texelPerUnit));
	LMAxisT = bestAxis2 * (1.0f / (bestYd / p_texelPerUnit));

	// Now allocate a position for the surface in the dynamic lightmap
	if (p_pixelStorageMgr)
	{
		unsigned x, y;
		if (!p_pixelStorageMgr->AllocateBlock(p_widthOut, p_heightOut, x, y))
		{
			p_pixelStorageMgr->Initialize(DYNAMICLIGHTMAPDIMENSION, DYNAMICLIGHTMAPDIMENSION);
			p_pixelStorageMgr->AllocateBlock(p_widthOut, p_heightOut, x, y);
			if (p_dynLMID)
				*p_dynLMID = *p_dynLMID + 1;
		}

		dynLMX = x;
		dynLMY = y;
		dynLMXf = float(x);
		dynLMYf = float(y);
		if (p_dynLMID)
			dynLMID = *p_dynLMID;

		//dynLMX = 0;
		//dynLMY = 0;
		//dynLMXf = 0;
		//dynLMYf = 0;
		//dynLMID = -1;	// Setting this removes all optimized buffer switching

		TRACE ("Lightmap coords: %li %li, width = %li, %li; ID = %li\n", dynLMX, dynLMY, p_widthOut, p_heightOut, dynLMID);
	}
	else
	{
		dynLMX = 0;
		dynLMY = 0;
		dynLMXf = 0.0f;
		dynLMYf = 0.0f;
		dynLMID = -1;
	}
}


