// Surface3D.h: interface for the CSurface3D class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SURFACE3D_H__258D3E63_4597_419E_BB97_A830BEBB9477__INCLUDED_)
#define AFX_SURFACE3D_H__258D3E63_4597_419E_BB97_A830BEBB9477__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\math\Vector3D.h"
#include "ColorRGB.h"
#include "APITexture.h"
#include "ModelVertex3D.h"
#include "SurfaceDecal.h"
#include "..\structs\LinkedList.h"
#include "LightSource.h"
#include "..\memory\PixelStorageMgr.h"

class CTexCoord2f
{
public:
	float s, t;

	void GetLS(float &p_s, float &p_t)
	{
		p_s = s;
		p_t = t;
	}

	inline float GetS(void) {return s;}
	inline float GetT(void) {return t;}
};

class CSurfaceDecalList;

// NOTE: In most cases, a CSurface3D is intended to be a flat plane,
//   but as long as no real planar operations are needed (lighting, decals), 
//   a surface can be assembled and sent to Render->RenderSurfaces, 
//   and it will work.  CTerrain::RenderProjectedTexture does this
// Anything that would make use of GetNormal() or anything it helps to derive
//   should require a flat plane
class CSurface3D  
{
public:
	CSurface3D();
	virtual ~CSurface3D();

	void Destroy(void);

private:
	unsigned id;		//  When needed, this equals the index the surface has
						//   in whatever array it is a part of

	// Vertex reference array
	int *vertexRefs;
	unsigned vertexRefQty;

	CVector3D normal;		// unit normal to the surface, facing outward
							//   (assumes clockwise definition)
	float planeK;			// K for planar ax+by+cz+K = 0 equation
	CVector3D *segUnits;	// unit length vector along segment's length
	CVector3D *segNormals;	// Normals pointing inward from each segment
							//   (assumes concave surface)
	float *segLengths;		// lengths of segments

	bool useColor;			//  Color and whether or not to use it
	CColorRGB color;		//    if false, use vertex colors, but
							//    if vertices are all the same color,
							//    setting this to true makes rendering
							//    faster

	BYTE alpha;				// alpha for the surface, normally 255

	CTexCoord2f *texCoords;	// Texture coordinates
	CAPITexture *texture;	// texture to use (NULL = none)
							//    This is a reference, registered with
							//    APITextures, so it will be destroyed there

	bool visible;			// Is the surface visible and collidable?

	bool mirror;			// Is it a mirror?  (important for portal map rendering)
	bool floor;				// Is it a floor? (important for projecting a floor texture)

	int lightmapID;			// Lightmap ID to use (-1 for none, otherwise index for
							//     array passed to RenderSurfaces)
	CTexCoord2f *texCoordsLM;	// Texture coordinates for the lightmap

	// Used to convert a point on the surface into a texture coordinate
	//   - allows spliced surfaces to maintain uniform texturing
	CVector3D	Tex1Origin;
	CVector3D	Tex1AxisS;
	CVector3D	Tex1AxisT;

	// Used to convert a point on the surface into a lightmap coordinate
	CVector3D	LMOrigin;
	CVector3D	LMAxisS;
	CVector3D	LMAxisT;

	unsigned dynLMX, dynLMY;	// Starting coords for off-screen lightmap
	float dynLMXf, dynLMYf;		// Same values in float
	int	dynLMID;				// ID of offscreen lightmap to use
								//  (this is jsut a way of keeping clusters of
								//    surfaces together, like radio buttons and a
								//    group ID)

public:
	void SetLMOrigin(CVector3D &p_vector) {LMOrigin = p_vector;}
	void SetLMAxisS(CVector3D &p_vector) {LMAxisS = p_vector;}
	void SetLMAxisT(CVector3D &p_vector) {LMAxisT = p_vector;}
	CVector3D & GetLMOrigin(void) {return LMOrigin;}
	CVector3D & GetLMAxisS(void) {return LMAxisS;}
	CVector3D & GetLMAxisT(void) {return LMAxisT;}

private:
	int decalListID;		// index for decal list (-1 if none, otherwise
							//    index of array passed to RenderSurfaces)

	CSurface3D	*opposingSurface;	// Surface opposite this one (important
									//   for decaling other side of an alpha surface)


	inline void CalculateLightmapArea(CModelVertex3D *p_vertices, float p_texelPerUnit, unsigned &p_widthOut, unsigned &p_heightOut)
	{
		CalculateLightmapArea(p_vertices, p_texelPerUnit, p_widthOut, p_heightOut, NULL, NULL);
	}
	void CalculateLightmapArea(CModelVertex3D *p_vertices, float p_texelPerUnit, unsigned &p_widthOut, unsigned &p_heightOut, CPixelStorageMgr *p_pixelStorageMgr, unsigned *p_dynLMID);
	bool GetSegmentIntersection(CVector3D &p_start, CVector3D &p_directionUnit, float p_segLength, bool p_outside, CModelVertex3D *p_vertices, CVector3D &p_intersectionPointOut, unsigned &p_segIDOut);

public:
	inline void Initialize(unsigned p_vertexQty)
	{
		Initialize(p_vertexQty, NULL);
	}
	// for compatibility with VB
	inline void Initialize(unsigned p_vertexQty, bool p_dummy, CAPITexture *p_texture)
	{
		Initialize(p_vertexQty, p_texture);
	}
	//inline void Initialize(unsigned p_vertexQty, char *p_textureName)
	//{
	//	Initialize(p_vertexQty, APITextures.GetTexture(p_textureName));
	//}
	void Initialize(unsigned p_vertexQty, CAPITexture *p_texture);

	inline unsigned GetID() { return id;}
	inline int GetLightmapID() {return lightmapID; }
	inline int GetDecalListID() {return decalListID; }

	inline unsigned GetVertexRefQty() {return vertexRefQty;}
	inline int * GetVertexRefs() {return vertexRefs;}
	inline int GetVertexRef(unsigned p_id) {return vertexRefs[p_id];}

	inline CVector3D & GetNormal() {return normal;}
	inline float GetPlaneK(void) {return planeK;}
	inline float CalculateYatXZ(CVector3D &p_pos)
	{
		return (-(normal.x*p_pos.x + normal.z*p_pos.z + planeK)/normal.y);
	}

	void SetTexture(CAPITexture *p_texture)
	{
		texture = p_texture;
	}
	void SetTexture(char *p_name);
	inline CAPITexture * GetTexture(void) {return texture;};

	void SetID(unsigned p_id) { id = p_id;}
	void SetLightmapID(int p_id) {lightmapID = p_id;}
	void SetDecalListID(int p_id) {decalListID = p_id;}
	void SetOpposingSurface(CSurface3D *p_surface) {opposingSurface = p_surface;}
	void SetVisible(bool p_visible) {visible = p_visible;}
	void SetNormal(CVector3D &p_vector) {normal = p_vector;}

	// Not sure if we need these
	//void SetFloor(bool p_floor) {floor = p_floor;}
	//inline bool IsFloor() {return floor;}

	void SetMirror(bool p_mirror) {mirror = p_mirror;}
	void SetAlpha(BYTE p_alpha) {alpha = p_alpha;}
	inline bool IsMirror() {return mirror;}
	inline BYTE GetAlpha() {return alpha;}
	inline bool IsVisible() {return visible;}

	inline bool UseSurfaceColor(void) {return useColor;}
	inline CColorRGB GetColor(void) {return color;}

	inline CSurface3D * GetOpposingSurface(void) {return opposingSurface;}

	void SetColor(BYTE p_red, BYTE p_green, BYTE p_blue)
	{
		useColor = true;
		color.SetColor(p_red, p_green, p_blue);
	}
	void SetColor(const CColorRGB &p_color)
	{
		useColor = true;
		color.SetColor(p_color);
	}

	inline CTexCoord2f * const GetTexCoords() {return texCoords;}
	inline CTexCoord2f * const GetTexCoordsLM() {return texCoordsLM;}

	// Texture coordinates for offscreen buffer
	// Defines origin for lightmap, since multiple surfaces
	//   can share the same offscreen buffer
	// if dynLMID = -1, then render routine assumes that surface
	//    uses its own
	inline unsigned GetDynLMX() {return dynLMX;}
	inline unsigned GetDynLMY() {return dynLMY;}
	inline float GetDynLMXf() {return dynLMXf;}
	inline float GetDynLMYf() {return dynLMYf;}
	inline int GetDynLMID(void) {return dynLMID;}

	inline void SetDynLMOrigin(unsigned p_x, unsigned p_y)
	{
		dynLMX = p_x;
		dynLMY = p_y;
		dynLMXf = float(p_x);
		dynLMYf = float(p_y);
	}

	void SetVertexRef(unsigned p_index, int p_vertexRefID);
	void SetVertexRef(unsigned p_index, int p_vertexRefID, float p_s, float p_t);
	// for compatibility with VB version
	inline void SetVertexRefTexture(unsigned p_index, int p_vertexRefID, float p_s, float p_t)
	{
		SetVertexRef(p_index, p_vertexRefID, p_s, p_t);
	}
	// After the vertex references are set up, this can be called
	void CalculateSurfaceData(CModelVertex3D *p_vertices);
	// And this after tex coord setup
	void CalculateTex1Axes(CModelVertex3D *p_vertices);

	void SetVertexTexCoords(unsigned p_index, float p_s, float p_t);
	void SetVertexLMCoords(unsigned p_index, float p_sLM, float p_tLM);

	bool PointInside(CVector3D &p_point, CModelVertex3D *p_vertices);
	bool SphereCollision(CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CModelVertex3D *p_vertices, CVector3D &p_collisionNormalOut);
	bool RayCollision(CVector3D &p_start, CVector3D &p_directionUnit, float &p_collisionDistanceOut, CModelVertex3D *p_vertices, CVector3D &p_collisionPointOut);
	bool PointInFront(CVector3D &p_point, CModelVertex3D *p_vertices)
	{
		return (((p_point - p_vertices[vertexRefs[0]].vertex) * normal) >= -0.0001f);
	}

	inline void PlaceDecal(CSurface3D &p_decalSurface, CModelVertex3D *p_decalVertices, CModelVertex3D *p_verticesThis, CSurfaceDecalList *p_decalLists)
	{
		PlaceDecal(p_decalSurface, p_decalVertices, p_verticesThis, p_decalLists, NULL);
	}
	void PlaceDecal(CSurface3D &p_decalSurface, CModelVertex3D *p_decalVertices, CModelVertex3D *p_verticesThis, CSurfaceDecalList *p_decalLists, CSurfaceDecalList *p_decalLists2);

	void CalculateTex1Coords(CVector3D &p_point, float &p_sOut, float &p_tOut);
	void CalculateTexLMCoords(CVector3D &p_point, float &p_sOut, float &p_tOut);

	bool SurfaceIntersection(CSurface3D &p_surfaceP, CModelVertex3D *p_verticesP, CModelVertex3D *p_verticesThis, CVector3D *p_vertOut, unsigned &p_vertexQtyOut);

	//////////////////////////////
	// Lightmap functions
	void ClearLightmap(CColorRGB p_color, CAPITexture **p_lightmaps);
	void ApplyRadiusLightLM(float p_x, float p_y, float p_radius, CColorRGB p_color, CAPITexture **p_lightmaps);
	inline void SetupLightmap(CModelVertex3D *p_vertices, CAPITexture **p_lightmaps, float p_texelPerUnit)
	{
		SetupLightmap(p_vertices, p_lightmaps, p_texelPerUnit, NULL, NULL);
	}
	void SetupLightmap(CModelVertex3D *p_vertices, CAPITexture **p_lightmaps, float p_texelPerUnit, CPixelStorageMgr *p_pixelStorageMgr, unsigned *p_dynLMID);
	void ApplyLightSourceLM(CLightSource *p_ls, CModelVertex3D *p_vertices, CAPITexture **p_lightmaps);
};

class CSurfaceDecal  
{
public:
	CSurfaceDecal();
	virtual ~CSurfaceDecal();

	CSurface3D surface;
	CModelVertex3D *vertices;
	unsigned vertexQty;

	void Initialize(unsigned p_vertexQty)
	{
		vertices = new CModelVertex3D [p_vertexQty];
		vertexQty = p_vertexQty;
	}

	inline CSurface3D & GetSurface(void) {return surface;}
	inline CModelVertex3D & GetVertex(unsigned p_id) {return vertices[p_id];}

	bool backside;	// Is this decal on the backside of an alpha surface?

	void Destroy(void);
};

class CSurfaceDecalListNode: public CLinkedListNode<CSurfaceDecal>
{
};

class CSurfaceDecalList: public CLinkedList<CSurfaceDecal>
{
public:
	virtual ~CSurfaceDecalList();
	void DestroyDecals(void);		// Destroys decals in the list
	void Delete (CLinkedListNode<CSurfaceDecal> *p_node)
	{
		// Before deleting, get rid of decal storage
		TRACE("Destroying a decal\n");
		p_node->Destroy();

		CLinkedList<CSurfaceDecal>::Delete(p_node);
	}
	void EmptyList(void)
	{
		// Before Emptying, destroy the decals
		TRACE("Destroying decals before emptying list\n");
		DestroyDecals();
		CLinkedList<CSurfaceDecal>::EmptyList();
	}

	CSurfaceDecalListNode * NewNode(void)
	{
		return NewNode(true);
	}
	CSurfaceDecalListNode * NewNode(bool p_addFirst)
	{
		return (CSurfaceDecalListNode *)CLinkedList<CSurfaceDecal>::NewNode(p_addFirst);
	}

	// I would like to put a destructor in the CLinkedListNode
	//   portion for CSurfaceDecal, but I had a problem getting NewNode
	//   to work.  Maybe I can just typecast?  And, the warning about
	//   the stack makes me nervous.
};

#endif // !defined(AFX_SURFACE3D_H__258D3E63_4597_419E_BB97_A830BEBB9477__INCLUDED_)
