// Terrain.h: interface for the CTerrainPoint class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TERRAINPOINT_H__13163905_41FD_4A49_B809_B9502F9CEB91__INCLUDED_)
#define AFX_TERRAINPOINT_H__13163905_41FD_4A49_B809_B9502F9CEB91__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\math\Vector3D.h"
#include "ColorRGB.h"
#include "APITexture.h"

class CTerrainPoint  
{
public:
	CTerrainPoint();
	virtual ~CTerrainPoint();

	float y;	// Height at this point (x,z are the terrain axes)

	CVector3D primaryNormal;	// Normal for triangle 1
	CVector3D secondaryNormal;	// Normal for triangle 2
	float pK;		// Constant for triangle 1 plane equation
	float sK;		// Constant for triangle 2 plane equation

	CVector3D	pointNormal;	// Unitized normal for the vertex itself
								//   - provides a more rounded lighting look

	BYTE		color;	// Color for this vertex - now a combination of
						//   CTerrain::Shade and illumination, not needed anymore
	BYTE		alpha1;	// Alpha1 for this vertex - could be more

	float illumination; // amount to add to shade to reach 255 (if shade is 32
						//   and illumination is 0.5, end result is 32 + (255-32) * 0.5 = 239
};

enum defTerrainRenderType
{
    TERRAINRENDER_NOTEXTURE = 0,
    TERRAINRENDER_SINGLETEXTURE,
    TERRAINRENDER_TWOTEXTUREBLENDALPHA		//   each vertex will have an alpha
                                            //   value that will determine which texture
                                            //   shows up more (0 = t1, 255 = t2)
                                            
                                            // Assume terrain will always be modulated by the
                                            //   color at each vertex, for now
};

// This class encompasses an entire chunk of renderable terrain

// The terrain in memory looks like this:
// (2,2) (1,2) (0,2)
// (2,1) (1,1) (0,1)
// (2,0) (1,0) (0,0)

// and a cell looks like this:
// -----
// |s /|
// | / |  primary/secondary triangles (each has its own normal)
// |/ p|
// ----o  <---- this is the point with the y value in the points array
//
// The left and topmost points of the terrain chunk only store y,
//   no normals (there are no squares there)

class CTerrain
{
public:
	CTerrain();
	virtual ~CTerrain();
	void Destroy();

	unsigned xQty;     // Dimensions of terrain (points, not cells - cells qty = points-1)
	unsigned zQty;

	float cellSize;  // World size of a cell

	BYTE shade;		// color value of shade (0-255), combined with illumination
					//   at point level to determine final color (white-based)

	float startS0, startT0;	// Where to start with textures
	float startS1, startT1;

	// Full bounds for the terrain (helps speed up collision detection and occlusion)
	// (not used yet)
	CVector3D highPoint;
	CVector3D lowPoint;

	CAPITexture *texture0;			// Texture to decal on terrain, if any
	float texCoordPerCell0;			// Amount to advance s and t per cell
	CAPITexture *texture1;			// Secondary texture to render (multitexturing)
	float texCoordPerCell1;			// Amount to advance s and t per cell
	defTerrainRenderType  textureOp;	// Method to render texture(s)
	// ?? Naturally we'll want to multitexture a detail on there to make the closer cells look nice

	CTerrainPoint *points;			// Vertices that make up the terrain

	void Initialize(unsigned p_xQty, unsigned p_zQty, float p_cellSize);
	inline void SetTexture0(CAPITexture *p_texture0, float p_texCoordPerCell0)
	{
		texture0 = p_texture0;
		texCoordPerCell0 = p_texCoordPerCell0;
	}
	inline void SetTexture1(CAPITexture *p_texture1, float p_texCoordPerCell1)
	{
		texture1 = p_texture1;
		texCoordPerCell1 = p_texCoordPerCell1;
	}
	inline void SetTextureOp(defTerrainRenderType p_textureOp) {textureOp = p_textureOp;}
	CVector3D GetNormalAtXZ(float p_x, float p_z);
	float CalculateYAtXZ(float p_x, float p_z);

	void SetY(unsigned p_xIndex, unsigned p_zIndex, float p_y) { points[p_xIndex+p_zIndex*xQty].y = p_y; }
	void SetColor(unsigned p_xIndex, unsigned p_zIndex, BYTE p_color) { points[p_xIndex+p_zIndex*xQty].color = p_color; }
	void SetAlpha1(unsigned p_xIndex, unsigned p_zIndex, BYTE p_alpha) { points[p_xIndex+p_zIndex*xQty].alpha1 = p_alpha; }
	void SetIllumination(unsigned p_xIndex, unsigned p_zIndex, float p_illumination) { points[p_xIndex+p_zIndex*xQty].illumination = p_illumination; }

	float GetY(unsigned p_xIndex, unsigned p_zIndex) { return (points[p_xIndex+p_zIndex*xQty].y); }
	BYTE GetColor(unsigned p_xIndex, unsigned p_zIndex) { return points[p_xIndex+p_zIndex*xQty].color; }
	CVector3D GetPointNormal(unsigned p_xIndex, unsigned p_zIndex) { return (points[p_xIndex+p_zIndex*xQty].pointNormal); }
	BYTE GetAlpha1(unsigned p_xIndex, unsigned p_zIndex) { return (points[p_xIndex+p_zIndex*xQty].alpha1); }
	CTerrainPoint * GetPoints(void) {return points;}

	inline void CalculateSurfaceData(void)
	{
		CalculateSurfaceData(true);
	}
	void CalculateSurfaceData(bool p_calcAlpha);

	void LoadFromBMP(char *p_BMPfileName, float p_cellSize, float p_minY, float p_maxY);
	bool RayCollision(CVector3D &p_start, CVector3D &p_directionUnit, float &p_collisionDistanceOut, CVector3D &p_collisionPointOut);

	inline void RenderProjectedTexture(void *p_render, float p_x, float p_z, float p_radius, float p_rotation, CAPITexture *p_texture)
	{
		RenderProjectedTexture(p_render, p_x, p_z, p_radius, p_rotation, p_texture, NULL, 255);
	}
	void RenderProjectedTexture(void *p_render, float p_x, float p_z, float p_radius, float p_rotation, CAPITexture *p_texture, CColorRGB *p_modColor, BYTE p_alpha);
};

#endif // !defined(AFX_TERRAINPOINT_H__13163905_41FD_4A49_B809_B9502F9CEB91__INCLUDED_)
