// PtlMap.cpp: implementation of the CPtlMap class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PtlMap.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

unsigned int CPtlMap::renderID = 0;

CPtlMap::CPtlMap()
{
	rooms = NULL;
	portals = NULL;
	vertices = NULL;

	roomQty = 0;
	portalQty = 0;
	vertexQty = 0;

	mirrorDepth = 0;	// This should always return to zero after
						//   Render finishes
}

CPtlMap::~CPtlMap()
{
	Destroy();
}

void CPtlMap::Destroy(void)
{
	TRACE ("Destroying portal map\n");

	if (rooms) delete [] rooms;
	if (portals) delete [] portals;
	if (vertices) delete [] vertices;

	rooms = NULL;
	portals = NULL;
	vertices = NULL;

	roomQty = 0;
	portalQty = 0;
	vertexQty = 0;

	lightSources.EmptyList();
	renderObjects.EmptyList();
}

void CPtlMap::Initialize(unsigned p_vertexQty, unsigned p_roomQty, unsigned p_portalQty)
{
	vertexQty = p_vertexQty;
	if (vertexQty)
		vertices = new CModelVertex3D[vertexQty];

	roomQty = p_roomQty;
	if (roomQty)
	{
		rooms = new CPtlRoom[roomQty];

		unsigned int i;
		for (i=0; i<roomQty; i++)
			rooms[i].SetID(i);
	}

	portalQty = p_portalQty;
	if (portalQty)
		portals = new CPtlPortal[portalQty];
}

void CPtlMap::CalculateSurfaceData(float p_targetVolume, float p_minDimension)
{
	unsigned i;

	// Get bounding volume for map and initialize octree
	minPoint.SetVec(100000,100000,100000);
	maxPoint.SetVec(-100000,-100000,-100000);
	for (i=0; i<vertexQty; i++)
	{
		if (vertices[i].vertex.x < minPoint.x) minPoint.x = vertices[i].vertex.x;
		if (vertices[i].vertex.y < minPoint.y) minPoint.y = vertices[i].vertex.y;
		if (vertices[i].vertex.z < minPoint.z) minPoint.z = vertices[i].vertex.z;

		if (vertices[i].vertex.x > maxPoint.x) maxPoint.x = vertices[i].vertex.x;
		if (vertices[i].vertex.y > maxPoint.y) maxPoint.y = vertices[i].vertex.y;
		if (vertices[i].vertex.z > maxPoint.z) maxPoint.z = vertices[i].vertex.z;
	}
	octree.SetBounds(minPoint, maxPoint);

	// Calculate data on room surface and split octree
	for (i=0; i<roomQty; i++)
        rooms[i].CalculateSurfaceData(vertices, octree, p_targetVolume, p_minDimension);
    
    for (i=0; i<portalQty; i++)
        portals[i].CalculateData(vertices);

	// Now get lists of all cells each room is touching
	for (i=0; i<roomQty; i++)
        rooms[i].DetermineCellsTouched(octree);
}

// This is for those alpha blended surfaces that are opposite to
//   each other, to allow decals to be applied to both when
//   a decal is placed on one (allows decals to appear on the opposite
//   side as well, by placing one on each, a front one and a 'backside' one)
void CPtlMap::SetTwoOpposingSurfaces(unsigned p_roomID, unsigned p_surfaceID, unsigned p_opposingRoomID, unsigned p_opposingSurfaceID)
{
	// Watch out for these &'s, performing a & on a returned & might
	//   not provide the pointer value
    rooms[p_roomID].GetSurface(p_surfaceID).SetOpposingSurface(
        rooms[p_opposingRoomID].GetSurfacePtr(p_opposingSurfaceID));
    rooms[p_roomID].SetOpposingRoom(p_surfaceID, p_opposingRoomID);
    
    rooms[p_opposingRoomID].GetSurface(p_opposingSurfaceID).SetOpposingSurface(
        rooms[p_roomID].GetSurfacePtr(p_surfaceID));
    rooms[p_opposingRoomID].SetOpposingRoom(p_opposingSurfaceID, p_roomID);
}

/*
' Current rendering algorithm:
'
' 1. figure out which room we are in
' 2. Render that room
'
' Rendering a room:
'    If room was already rendered, return
'    For each unused portal visible in the current frustum
'        Create a new frustum, merging current with the portal
'        If the frustum is valid, render all rooms attached to
'           that portal

' This algorithm has a couple problems:
' ?? Figuring out which room we are in is expensive
'      - a BSP algorithm could be used for this
' * We may have circumstances where different frustums could
'    actually use different portions of the same portal, and
'    one frustum may see a distant portal when the other frustum
'    did not.  This means sometimes distant rooms may not be
'    rendered when they should.
' So, improve the algorithm:

' ?? Still need to do this yet
' * Keep track of which room we are in and tell the algorithm
'    which room to start in.  It's easy enough to tell which
'    portals we pass through as we move.  If we need to figure
'    out which room we are in, that should be done elsewhere,
'    not during rendering.  A BSP tree might be best for this
'    purpose if we aren't tracking how we move through rooms
' - Each portal can only have two adjacent rooms (front and back)
'     Don't simply render all adjacent rooms, pick which one
'     we should render based on the viewpoint's position relative
'     to the portal, always working away - done
' * Just because we've already rendered a room doesn't mean we
'     can't still check its portals for visibility, and we'll
'     need to in the event a later frustum penetrates the same room
' - Portals shouldn't be flagged as used, either, since a single
'     portal can be hit by multiple frustums - done
' * Ultimately, we should only be rendering the visible geometry,
'     but that doesn't change the methodology of our flagging.
' * If we do allow rendering of the same room more than once (either
'    by just redrawing, or only allowing visible geometry), we
'    should make sure dynamic lightmaps for that room are calculated
'    only once
' * Visible geometry will need to be applied to decals as well
'     This is especially true for alpha decals that would darken
'     themselves is totally drawn more than once
' Does this prevent infinite loops?  I believe so.
' And there is a circumstance where we want to save the room
'    rendering for the very last - when a portal has an alpha
'    blended surface on it.  So when we're done rendering one
'    portal in a room, we'll want to make sure we've first rendered
'    all other possible ways in which farther rooms from that
'    one will be rendered.  This is one case where rendering
'    only visibile geometry would help.
' Even if a map is built entirely out of reshaped cubes, we can still have this
'    problem.  Imagine a room with a thin post down the center, and
'    a doorway at the other end.  The post cuts the door in half in
'    our view.  The rendering would take care of one side, then the
'    other (a portal on each side of the post).
'    But if the doorway has an alpha surface, the first time the
'    far rooms are rendered, the alpha surface would be drawn and
'    could be corrupted by what's drawn on the other side of the post, especially
'    if a far portal is visible on one side of the post but not the other

' New algorithm:
'    For a given subparse:
'    If a starting room is given, use that, otherwise
'       figure out what room the viewpoint is in (room comparison (slow)
'       or BSP)
'    If this room isn't in the room list yet, add it after the
'       last room parsed (this list will be used to prevent rendering
'       a room twice, and ensure that everything is drawn in the
'       proper order - to make sure alpha surface properly blend
'       against farther surfaces)
'    For each portal the room is adjacent to, see if it is visible
'       in the current frustum
'       Test mirror portals first, then the normal portals
'          It's important to render mirror surfaces first
'          so that farther mirrors can't corrupt the depth buffer
'          before closer mirrors are rendered
'       If portal is visible,
'           Create a new frustum, merging the current and the
'              visible portal
'           If it's a mirror, stencil and planeclip the mirror portal
'               spliced by the new frustum (splicing necessary to prevent
'               overdraw, ie alpha accumulation, in case the mirror is parsed through
'               another process path, and stenciling is necessary
'               to keep mirror images from cutting into non-mirror
'               areas, such as a sideways hall in the mirror cutting into
'               a hallway going deeper beside the mirror), flip the frustum and
'               call a subrender, starting in this room, then
'               restore the stencil back to what it was before
'               Track mirror depth, don't go too deep
'               Also, set up a clipping plane on the mirror
'               that prevents geometry from extending out of the surface
'               (save the clipping plane at each mirror level and restore them
'                on return)
'           If it isn't a mirror, parse into the room on the
'               other side of the portal
'    When all parsing is done,
'       Depth mask all mirror portals in the room list front
'          to back, to prevent farther surfaces from cutting into
'          mirrored images
'       Render the control object if we are rendering a mirror image
'       Render the room list back to front
'       ?? Still need to figure out the right way to render
'          game objects, probably draw the objects in a room,
'          then the room

' This routine represents one instance of drawing EVERYTHING
' We need it this way because it's possible to render mirrors, and
'   each mirror will call this routine
*/
void CPtlMap::Render(IRender *p_render, CPtlFrustum &p_portalFrustum, DWORD p_elapsedTimeMS, DWORD p_currentTime, int p_startingRoomID, bool p_postparseRender)
{
	// Don't use statics here since it can be called recursively
	static unsigned int	renderID = 0;	// Well, this one is ok, it's incremented
										//   after this routine can't be called again

    CLinkedList<CInteger> roomList;  // List of rooms to render

	LPINTNODE *roomsRendered;
	roomsRendered = new LPINTNODE[roomQty];	// Has this room been set to render already?
        
    // Reset rendered flags
    unsigned i;
	// ?? This might be faster with a memset to zero
    for (i=0; i<roomQty; i++)
        roomsRendered[i] = NULL;
	roomList.EmptyList();	// Just in case
    
    // Find out which room we're in
    // ?? We could limit which rooms we test based on sorting volumes, or perhaps,
    //   track where the viewpoint was last time?
    if (p_startingRoomID != -1)
        RenderRoom(p_startingRoomID, p_render, p_portalFrustum, p_postparseRender, roomsRendered, roomList.GetHeader(), -1, p_elapsedTimeMS, p_currentTime);
    else
	{
		//TRACE(">>>>> BEGIN NEW RENDER <<<<<<\n");
        // Which room are we in?
		// ?? This is VERY slow, better to know what room we're in
		//   before we start (BSP)
		// Here we are using the octree now, checking all rooms touching the
		//   cell that the Viewpoint is in - not as fast as a BSP but still
		//   faster than testing all
		int roomID = RoomOfPoint(p_portalFrustum.GetViewpoint());
		if (roomID != -1)
		{
	        RenderRoom(roomID, p_render, p_portalFrustum, p_postparseRender, roomsRendered, roomList.GetHeader(), -1, p_elapsedTimeMS, p_currentTime);
		}
/*        for (i=0; i<roomQty; i++)
            if (rooms[i].PointInRoom(p_portalFrustum.GetViewpoint(), vertices))
			{
                // Render room - this kicks off everything else
                RenderRoom(i, p_render, p_portalFrustum, p_postparseRender, roomsRendered, roomList.GetHeader(), -1, p_elapsedTimeMS, p_currentTime);
                i = roomQty - 1;
			}*/
	}
    
    // We are done, render the rooms according to p_postParse
    LPINTNODE node1;

	renderID++;			// Establish ID for rendering, render objects not yet set to this value

    CColorRGB renderColor;
    
	// Object rendering is handled at the room level now
/*    if (mirrorDepth > 0)
	{
        // Render the control object if we're dealing with reflections
        renderColor.SetColor(lightSources.CalculateRenderColor(controlObject.orient.p, ambientColor));
        p_render->PushMatrix();
        p_render->TransformWorld(controlObject.orient);
        p_render->Scalef(controlObject.scalef.x, controlObject.scalef.y, controlObject.scalef.z);
        p_render->RenderModelColor(controlObject.modelRef, &renderColor);
        p_render->PopMatrix();
    }
    // Render the rest of the objects always
    // NOTE: Alpha objects should be rendered after everything else
    //    Same for alpha surfaces
    // ??
    renderColor.SetColor(lightSources.CalculateRenderColor(testObject.orient.p, ambientColor));
    p_render->PushMatrix();
    p_render->TransformWorld(testObject.orient);
    p_render->Scalef(testObject.scalef.x, testObject.scalef.y, testObject.scalef.z);
    p_render->RenderModelColor(testObject.modelRef, &renderColor);
    p_render->PopMatrix();
	//char out[50];
	//TRACE("Render color: %s\n", renderColor.OutStr(out));*/
    
	// Render the doors (this is ok as long as the doors aren't alpha, but this will change)
	// We should only draw the doors for which rooms are visible, or adjacent
	//   ones where the doors are visible
	// This is handled at the room level using the octree now
	/*
    for (i=0; i<portalQty; i++)
        portals[i].RenderDoors(p_render, &lightSources, renderID);
	*/

	CLightSourceRefList lights;

	// For each room about to be rendered, animate and place their particles
	// Order doesn't matter
    node1 = roomList.GetFirst();
	if (node1)
	while (node1->next != NULL)
	{
		rooms[node1->value].AnimateParticles(octree, rooms, vertices, p_elapsedTimeMS, p_currentTime);

		node1 = node1->next;
	}

    // Now render the rooms
    //If mirrorDepth = 0 Then GoTo leave  ' just a test
    if (p_postparseRender)
	{
        // Back to front
        node1 = roomList.GetLast();
        
		if (node1)
        while (node1->prev != NULL)
		{
			// Assemble a light list for the room
			lights.EmptyList();
			CLightSource::currentRenderID++;
			CLinkedListNode<COctCellRef> *cnode1;
			CLinkedListNode<CLPVoid> *vnode1;
			cnode1 = rooms[node1->value].GetCellsTouched().GetFirst();
			if (cnode1)
			while (cnode1->next != NULL)
			{
				vnode1 = cnode1->cellRef->lpVoidList2.GetFirst();
				if (vnode1)
				while (vnode1->next != NULL)
				{
					if (((CLightSource *)(vnode1->lpVoid))->renderID != CLightSource::currentRenderID)
					{
						((CLightSource *)(vnode1->lpVoid))->renderID = CLightSource::currentRenderID;
						lights.NewNode()->ref = (CLightSource *)vnode1->lpVoid;
					}

					vnode1 = vnode1->next;
				}

				cnode1 = cnode1->next;
			}

            rooms[node1->value].Render(vertices, p_render, p_elapsedTimeMS, p_currentTime, mirrorDepth, renderID, &lights, &ambientColor);
            //Debug.Print "Render room " & node1.roomID
        
            node1 = node1->prev;
		}
	}
    else
	{
        // Front to back
        node1 = roomList.GetFirst();
        
		if (node1)
        while (node1->next != NULL)
		{
			// Assemble a light list for the room
			lights.EmptyList();
			CLightSource::currentRenderID++;
			CLinkedListNode<COctCellRef> *cnode1;
			CLinkedListNode<CLPVoid> *vnode1;
			cnode1 = rooms[node1->value].GetCellsTouched().GetFirst();
			if (cnode1)
			while (cnode1->next != NULL)
			{
				vnode1 = cnode1->cellRef->lpVoidList2.GetFirst();
				if (vnode1)
				while (vnode1->next != NULL)
				{
					if (((CLightSource *)(vnode1->lpVoid))->renderID != CLightSource::currentRenderID)
					{
						((CLightSource *)(vnode1->lpVoid))->renderID = CLightSource::currentRenderID;
						lights.NewNode()->ref = (CLightSource *)vnode1->lpVoid;
					}

					vnode1 = vnode1->next;
				}

				cnode1 = cnode1->next;
			}

            rooms[node1->value].Render(vertices, p_render, p_elapsedTimeMS, p_currentTime, mirrorDepth, renderID, &lights, &ambientColor);
        
            node1 = node1->next;
        }
    }

	if (roomsRendered)
		delete [] roomsRendered;
}

void CPtlMap::RenderRoom(int p_roomID, IRender *p_render, CPtlFrustum &p_portalFrustum, bool p_postparseRender, LPINTNODE *p_roomsRendered, LPINTNODE p_lastNode, int p_lastRoomID, DWORD p_elapsedTimeMS, DWORD p_currentTime)
{
	// We'll definitely be rendering this room, just need to parse the
	//   rest of the portals first

	// Don't use statics here since it can be called recursively
    unsigned *adjacentPortalIDs; // Reference to array
	unsigned adjacentPortalQty;	// Number of entries in the array
        
    unsigned i;
    
    if (p_roomID == -1) return;

    CPtlFrustum newPortalFrustum;	// Frustum used by visible portals on next parse
    CPtlFrustum flippedFrustum;		// Special frustum used by mirrors
    
	// For rendering mirror surface
    CVector3D splicedPortalVerts[64];
	unsigned splicedPortalVertQty;

	//TRACE ("....Rendering room %i\n", p_roomID);

    // If room hasn't been rendered yet, add it to the list
    // Room ID's are always inserted right after the last room
    //    parsed.  That way, we make sure all rooms are
    //    rendered once and in the proper order at the end, even if we
    //    parse through the same room twice along the way
    if (!p_roomsRendered[p_roomID])
	{
		p_roomsRendered[p_roomID] = CLinkedList<CInteger>::NewNodeDetached();

		p_roomsRendered[p_roomID]->value = p_roomID;

		p_roomsRendered[p_roomID]->prev = p_lastNode;
		p_roomsRendered[p_roomID]->next = p_lastNode->next;
		p_lastNode->next->prev = p_roomsRendered[p_roomID];
		p_lastNode->next = p_roomsRendered[p_roomID];
	}

	// Debug
//	if (mirrorDepth == 1)
//		mirrorDepth = mirrorDepth;

	// Any portal to check?
    if (rooms[p_roomID].GetAdjacentPortalQty() == 0) return;

	// Get adjacentPortalIDs and how many
	adjacentPortalIDs = rooms[p_roomID].GetAdjacentPortalIDs();
	adjacentPortalQty = rooms[p_roomID].GetAdjacentPortalQty();

	// Loop once, testing mirrors, then loop again testing non-mirrors
	bool mirrorTest = true;
 	for (i=0; i<adjacentPortalQty; i++)
	{
       if (portals[adjacentPortalIDs[i]].IsMirror() == mirrorTest)
		{
		   // Debug
		   //if ((adjacentPortalIDs[i] == 1) && (mirrorDepth == 1))
			//   i=i;

            if (!portals[adjacentPortalIDs[i]].IsClosed())
            if (portals[adjacentPortalIDs[i]].IsVisible(p_portalFrustum, vertices, p_roomID))
			{
				// If portal visible, make a new frustum out of the old one and
				//   the one formed by the portal, proceed to the room beyond it,
				//   render the room, and parse there (recursively call
				//   this routine)
				newPortalFrustum.Initialize();
                if (portals[adjacentPortalIDs[i]].MergeFrustums(p_portalFrustum, newPortalFrustum, vertices, p_roomID))
				{
	                if (portals[adjacentPortalIDs[i]].IsMirror())
					{
						// Render the mirror (will call CPtlMap::Render again
						//    as a whole new render, so avoid function statics!!)
						if (mirrorDepth < MAXMIRRORDEPTH)
						{
							// Increment mirror depth (even 3 will kill
							//   the CPU)
							mirrorDepth = mirrorDepth++;

                            // Draw the portal with a stencil to limit what can be drawn
                            if (mirrorDepth == 1)
                                p_render->SetStencilEnable(true);

							// Start out with just stenciling the surface
                            p_render->SetStencilIncrement(mirrorDepth - 1);
                            p_render->EnableColorWrite(false);
                            p_render->EnableDepthWrite(false);
                            p_render->EnableCulling(false);
                            // Store the vertices we calculated so we don't have to recalculate
                            portals[adjacentPortalIDs[i]].RenderSpliced(p_render, vertices, p_portalFrustum, splicedPortalVerts, splicedPortalVertQty);
							// Set up for normal rendering (except for stenciling)
                            p_render->EnableCulling(true);
                            p_render->EnableDepthWrite(true);
                            p_render->EnableColorWrite(true);
                            
                            // Was the splice successful?
                            if (splicedPortalVertQty > 0)
							{
                                // Set stencil to only accept pixels that
                                //   are written in the area of the mirror surface
                                p_render->SetStencilTest(mirrorDepth);
                            
                                // Set clip plane variables
                                // This will actually work since we are flipping the viewpoint before we apply it
                                //   Otherwise we would reverse the normal
                                mirrorNormals[mirrorDepth - 1] = portals[adjacentPortalIDs[i]].GetNormal();
                                mirrorPoints[mirrorDepth - 1] = portals[adjacentPortalIDs[i]].GetVertex(0, vertices);
                                
                                // Make a new frustum by flipping newFrustum in the mirror
                                newPortalFrustum.ReflectOnPlane( 
                                    portals[adjacentPortalIDs[i]].GetVertex(0, vertices),
                                    portals[adjacentPortalIDs[i]].GetNormal(),
									flippedFrustum);
                                
                                // Flip the API viewpoint in the mirror
                                COrient3D flippedVP;
                                p_render->PushMatrix();
                                // Get API viewpoint
                                // Flip it
                                flippedVP = p_render->GetViewpoint().ReflectOnPlane(
                                    portals[adjacentPortalIDs[i]].GetVertex(0, vertices),
                                    portals[adjacentPortalIDs[i]].GetNormal(),
                                    REFLECTEDVPADJUST);
                                // Apply it
                                p_render->LoadIdentity();
                                p_render->Rotate(180, 0, 1, 0);
                                p_render->TransformViewpoint(flippedVP);
                                
                                // Establish a clip plane on the mirror so that
                                //  no geometry cuts from the mirror world into the real one
                                //   (combined with stenciling, this prevents everything - the
                                //    clip plane is necessary for polygons that extend out of the
                                //    mirror)
                                // We are doing this here because the modelview matrix
                                //   at this point is what it will be when/if we restore
                                //   this clip plane after a deeper mirror is parsed
                                //Debug.Print "Clip plane for mirror " & mirrorDepth - 1 & mirrorNormals(mirrorDepth - 1).OutputString & ";" & mirrorPoints(mirrorDepth - 1).OutputString
                                p_render->SetClipPlane(mirrorNormals[mirrorDepth - 1], mirrorPoints[mirrorDepth - 1]);
                                
                                // Toggle CW culling
                                p_render->SetCullCCW(!p_render->GetCullCCW());
                                
                                // Start a new render in the room with the mirror
                                // Renders rooms, control object and other objects in the portal
                                //    map lists
                                Render(p_render, flippedFrustum, p_elapsedTimeMS, p_currentTime, p_roomID, p_postparseRender);
                                                            
                                // Toggle CW culling back
                                p_render->SetCullCCW(!p_render->GetCullCCW());
                                
                                // Restore API viewpoint
                                p_render->PopMatrix();
                                //Call p_render.TransformViewpoint(flippedVP)
                            
                                if (mirrorDepth > 1)
								{
                                    // Restore clip plane for the previous mirror
                                    //Debug.Print "Restoring clip plane for mirror " & mirrorDepth - 1 & mirrorNormals(mirrorDepth - 1).OutputString & ";" & mirrorPoints(mirrorDepth - 1).OutputString
                                    p_render->SetClipPlane(mirrorNormals[mirrorDepth - 2], mirrorPoints[mirrorDepth - 2]);
								}
                                else
                                    p_render->DisableClipPlane();

                                // We're about to render the surface decorating the mirror, so
								//   we need to prepare a list of light source references now
								// Collect all lights touching the same cells this room is
								// ??
								CLightSourceRefList lights;
								CLinkedListNode<COctCellRef> *cnode1;
								CLinkedListNode<CLPVoid> *vnode1;

								lights.EmptyList();
								CLightSource::currentRenderID++;

								cnode1 = rooms[p_roomID].GetCellsTouched().GetFirst();
								if (cnode1)
								while (cnode1->next != NULL)
								{
									vnode1 = cnode1->cellRef->lpVoidList2.GetFirst();
									if (vnode1)
									while (vnode1->next != NULL)
									{
										if (((CLightSource *)vnode1->lpVoid)->renderID != CLightSource::currentRenderID)
										{
											((CLightSource *)vnode1->lpVoid)->renderID = CLightSource::currentRenderID;
											lights.NewNode()->ref = (CLightSource *)vnode1->lpVoid;
										}

										vnode1 = vnode1->next;
									}

									cnode1 = cnode1->next;
								}

                                // Undo the stencil back to what it was before
                                // Ignoring depth here in case anything from the rendered mirror
                                //    actually extends out of it, because we MUST clear the stencil
                                p_render->EnableCulling(false);
                                p_render->SetDepthTestFunc(APIDEPTHTESTFUNC_ALWAYS);
                                p_render->SetStencilDecrement(mirrorDepth, true);
                                // Use the vertices we calculated from RenderSpliced above
                                // We must re-render the polygon we calculated to be sure
                                //    we are fully restoring the stencil (re-rendering the
                                //    entire mirror surface sometimes misses a pixel or two,
                                //    apparently there is truncation when splicing)
                                // Here we are erasing the stencil, writing the
                                //   depth mask, and texturing the visible surface
                                //   if there is one.  If it has decals, we'll render
                                //   those later, I guess (still have to deal with
                                //   z fighting with those, but I'll displace them
                                //   forward rather than inward)
                                portals[adjacentPortalIDs[i]].RenderVertexArray(p_render, splicedPortalVerts, splicedPortalVertQty, 
									true, rooms[p_roomID].GetLightmaps(), &lights);
                                p_render->SetDepthTestFunc(APIDEPTHTESTFUNC_LESSEQUAL);
                                p_render->EnableCulling(true);
                                
                                if (mirrorDepth > 1)
                                    p_render->SetStencilTest(mirrorDepth - 1);
                                else
                                    p_render->SetStencilEnable(false);
                            }

                            // When we are completely done, mirrorDepth should be 0 again
                            mirrorDepth = mirrorDepth--;
                            if ((mirrorDepth == 0) && (splicedPortalVertQty == 0))
							{
                                // Make sure stencil is disabled in case
                                //   first splice failed
                                p_render->SetStencilEnable(false);
							}
                            
						}	// within mirrorDepth
					}	// is a mirror
					else
					{
                        // Render the room on the other side of the portal now (recursive)
                        //if (portals[adjacentPortalIDs[i]].PointInFront(newPortalFrustum.GetViewpoint(), vertices))
						if (portals[adjacentPortalIDs[i]].GetFrontRoomID() == p_roomID)
						{
                            // Guard against infinite loop from being right on a portal  (allowing dp=0 in IsVisible())
                            if (portals[adjacentPortalIDs[i]].GetBackRoomID() != p_lastRoomID)
                                RenderRoom(portals[adjacentPortalIDs[i]].GetBackRoomID(), p_render, newPortalFrustum, p_postparseRender, p_roomsRendered, p_roomsRendered[p_roomID], p_roomID, p_elapsedTimeMS, p_currentTime);
						}
                        else
						{
                            // Guard against infinite loop from being right on a portal (allowing dp=0 in IsVisible())
                            if (portals[adjacentPortalIDs[i]].GetFrontRoomID() != p_lastRoomID)
                                RenderRoom(portals[adjacentPortalIDs[i]].GetFrontRoomID(), p_render, newPortalFrustum, p_postparseRender, p_roomsRendered, p_roomsRendered[p_roomID], p_roomID, p_elapsedTimeMS, p_currentTime);
						}
					}	// is a normal portal
				}	// merge frustums successful

			}	// portal is visible
			else
			{
			   //if ((p_roomID == 1) && (adjacentPortalIDs[i] == 1) && (mirrorDepth == 1))
			//	::MessageBox(NULL, "Failed!", "Fail!", MB_OK);
			}
		}	// match mirror type for test

		// Test the non-mirrors now
		if ((i==(adjacentPortalQty - 1)) && (mirrorTest))
		{
			// Reset loop
			i=-1;	// This will set i to 0 when the loop continues
			mirrorTest = false;
		}
	}

}

void CPtlMap::AnimateDoors(long p_elapsedTimeMS)
{
	unsigned i;
	for (i=0; i<portalQty; i++)
		portals[i].AnimateDoors (p_elapsedTimeMS);
}

// NOTE: If the collision occurs with a door, the normal, surface and collision
//   point will be in object space, not world space
// Room surfaces are world space
bool CPtlMap::RayCollide(CVector3D &p_start, CVector3D &p_directionUnit, float &p_distanceOut, CVector3D &p_collisionPointOut, CSurface3D **p_collisionSurfaceOut, CPtlRoom **p_collisionRoomOut, CPtlDoor **p_collisionDoorOut)
{
    // Should start in current room and progress through portals
    // But for now, we'll do it the easy way
    // ??
    unsigned i;
	bool result = false;
    for (i=0; i<roomQty; i++)
	{
        if (RayCollision(rooms[i].GetSurfaces(), rooms[i].GetSurfaceQty(), vertices, p_start, p_directionUnit, p_distanceOut, p_collisionPointOut, p_collisionSurfaceOut))
		{
            *p_collisionRoomOut = &rooms[i];
            *p_collisionDoorOut = NULL;
			result = true;
		}
	}
    
    // We might have a surface collision at this point, but we still
    //   have to check doors.  If we get a collision there,
    //   then the door collision comes first
    for (i=0; i<portalQty; i++)
	{
        if (portals[i].RayCollisionWithDoor(p_start, p_directionUnit, p_distanceOut, p_collisionSurfaceOut, p_collisionPointOut, p_collisionDoorOut))
		{
            *p_collisionRoomOut = NULL;
            result = true;
		}
    }
    
    // Correct algorithm:
    // Figure out which room the start point is in (BSP)
    // Check for collisions on any collidable surfaces - if found leave
    // Otherwise check for ray to collide with any non-mirror portals
    //   For each found, start algorithm over again with the room on the
    //   other side
    // Check closest portals first (if that's possible, or necessary)
    //    I believe only one portal if any should satisfy it

	return result;
}

bool CPtlMap::SphereCollision(CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CSurface3D **p_collisionSurfaceOut, CVector3D &p_collisionWorldNormalOut, CPtlRoom **p_collisionRoomOut, CPtlDoor **p_collisionDoorOut)
{
//    unsigned i;
    
    // NOTE: p_collisionWorldNormalOut will be world space, not object!!!
    // To get the object space normal (door), use p_collisionSurface.GetNormal
    
    bool result = false;

	// Get a list of every cell the sphere would pass through on its path
	static CVector3D sminPoint;
	static CVector3D smaxPoint;
	sminPoint.x = p_pos.x - p_radius;
	sminPoint.y = p_pos.y - p_radius;
	sminPoint.z = p_pos.z - p_radius;
	smaxPoint.x = p_pos.x + p_radius;
	smaxPoint.y = p_pos.y + p_radius;
	smaxPoint.z = p_pos.z + p_radius;
	if (p_offset.x > 0)
		smaxPoint.x += p_offset.x;
	else
		sminPoint.x += p_offset.x;
	if (p_offset.y > 0)
		smaxPoint.y += p_offset.y;
	else
		sminPoint.y += p_offset.y;
	if (p_offset.y > 0)
		smaxPoint.z += p_offset.z;
	else
		sminPoint.z += p_offset.z;

	static COctCellRefList cellList;
	octree.CellsTouched(sminPoint, smaxPoint, cellList);

	static CVector3D tminPoint;
	static CVector3D tmaxPoint;
	float rad = float(p_radius * sqrt(2.0f));
	tminPoint.x = p_pos.x - rad;
	tminPoint.y = p_pos.y - rad;
	tminPoint.z = p_pos.z - rad;
	tmaxPoint.x = p_pos.x + rad;
	tmaxPoint.y = p_pos.y + rad;
	tmaxPoint.z = p_pos.z + rad;
	octree.CellsTouched(tminPoint, tmaxPoint, cellList);

	CPtlRoom::currentRenderID++;
	CPtlDoor::currentCollisionID++;

	CLinkedListNode<COctCellRef> *cnode1;
	CLinkedListNode<CInteger> *inode1;
	CLinkedListNode<CLPVoid> *vnode1;
	cnode1 = cellList.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		// Check rooms
		inode1 = cnode1->cellRef->intList.GetFirst();
		if (inode1)
		while (inode1->next != NULL)
		{
			if (rooms[inode1->value].renderID != CPtlRoom::currentRenderID)
			{
				rooms[inode1->value].renderID = CPtlRoom::currentRenderID;
				if (rooms[inode1->value].SphereCollisionRoom(vertices, p_pos, p_offset, p_radius, p_timeOut, p_collisionWorldNormalOut, p_collisionSurfaceOut))
				{
					result = true;
					*p_collisionRoomOut = &(rooms[inode1->value]);
					*p_collisionDoorOut = NULL;
				}
			}

			inode1 = inode1->next;
		}

		// Check doors
		vnode1 = cnode1->cellRef->lpVoidList1.GetFirst();
		if (vnode1)
		while (vnode1->next != NULL)
		{
			CPtlDoor *lpDoor = (CPtlDoor *)(vnode1->lpVoid);
			if (lpDoor->collisionID != CPtlDoor::currentCollisionID)
			{
				lpDoor->collisionID = CPtlDoor::currentCollisionID;

				if (lpDoor->IsClosed())
				{
					CVector3D newStart = lpDoor->currentOrient.TransformVectorToObjectSpace(p_pos - lpDoor->currentOrient.p);
					CVector3D newDirection = lpDoor->currentOrient.TransformVectorToObjectSpace(p_offset);
            				
					if (::SphereCollision(lpDoor->modelRef->GetSurfaces(), lpDoor->modelRef->GetSurfaceQty(), lpDoor->modelRef->GetVertices(), lpDoor->modelRef->GetVertexQty(), newStart, newDirection, p_radius, p_timeOut, p_collisionWorldNormalOut, p_collisionSurfaceOut))
					{
						result = true;

						// Transform collision normal back to space of p_pos and p_offset
						p_collisionWorldNormalOut =  lpDoor->currentOrient.TransformOffset(p_collisionWorldNormalOut);

						*p_collisionDoorOut = lpDoor;
						*p_collisionRoomOut = NULL;
					}
				}
			}

			vnode1 = vnode1->next;
		}

		cnode1 = cnode1->next;
	}

    /*for (i=0; i<roomQty; i++)
	{
		if (rooms[i].SphereCollisionRoom(vertices, p_pos, p_offset, p_radius, p_timeOut, p_collisionWorldNormalOut, p_collisionSurfaceOut))
		{
            result = true;
            *p_collisionRoomOut = &(rooms[i]);
            *p_collisionDoorOut = NULL;
        }
    }*/
    
    // Check all closed doors
    /*for (i=0;i <portalQty; i++)
	{
        if (portals[i].SphereCollisionWithDoors(p_pos, p_offset, p_radius, p_timeOut, p_collisionWorldNormalOut, p_collisionSurfaceOut, p_collisionDoorOut))
		{
			result = true;
            *p_collisionRoomOut = NULL;
		}
	}*/

	return result;
}

void CPtlMap::RenderFloorTexture(IRender *p_render, CVector3D &p_pos, float p_radius, float p_rotation, CAPITexture *p_texture, CColorRGB *p_modColor, BYTE p_modAlpha)
{
	// ?? We should figure out which room the texture actually touches and only
	//   attempt with those

	// Build a surface out of the texture
	// ??
	static CVector3D surfaceVertices[4];
	float a1 = float(cos((p_rotation+45.0f) * 3.14159266f/180.0f) * sqrt(2.0f) * p_radius);
	float a2 = float(sin((p_rotation+45.0f) * 3.14159266f/180.0f) * sqrt(2.0f) * p_radius);
	surfaceVertices[0].x = p_pos.x + a1;
	surfaceVertices[0].y = p_pos.y;
	surfaceVertices[0].z = p_pos.z + a2;
	surfaceVertices[1].x = p_pos.x - a2;
	surfaceVertices[1].y = p_pos.y;
	surfaceVertices[1].z = p_pos.z + a1;
	surfaceVertices[2].x = p_pos.x - a1;
	surfaceVertices[2].y = p_pos.y;
	surfaceVertices[2].z = p_pos.z - a2;
	surfaceVertices[3].x = p_pos.x + a2;
	surfaceVertices[3].y = p_pos.y;
	surfaceVertices[3].z = p_pos.z - a1;

	// Get a list of rooms this vertical column touches
	CPtlRoom::currentRenderID++;

	static COctCellRefList cellList;
	cellList.EmptyList();

	static CVector3D tminPoint;
	static CVector3D tmaxPoint;
	float rad = float(p_radius * sqrt(2.0f));
	tminPoint.x = p_pos.x - rad;
	tminPoint.y = p_pos.y - rad;
	tminPoint.z = p_pos.z - rad;
	tmaxPoint.x = p_pos.x + rad;
	tmaxPoint.y = p_pos.y + rad;
	tmaxPoint.z = p_pos.z + rad;
	octree.CellsTouched(tminPoint, tmaxPoint, cellList);

	CLinkedListNode<COctCellRef> *cnode1;
	CLinkedListNode<CInteger> *inode1;
	cnode1 = cellList.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		inode1 = cnode1->cellRef->intList.GetFirst();
		if (inode1)
		while (inode1->next != NULL)
		{
			if (rooms[inode1->value].renderID != CPtlRoom::currentRenderID)
			{
				rooms[inode1->value].renderID = CPtlRoom::currentRenderID;
				rooms[inode1->value].RenderFloorTexture(p_render, p_pos, surfaceVertices, p_texture, p_modColor, p_modAlpha, vertices);
			}

			inode1 = inode1->next;
		}

		cnode1 = cnode1->next;
	}

/*	unsigned i;
	for (i=0; i<roomQty; i++)
		rooms[i].RenderFloorTexture(p_render, p_pos, surfaceVertices, p_texture, p_modColor, p_modAlpha, vertices);
	*/
}

void CPtlMap::RenderLightFlares(IRender *p_render, CPtlFrustum &p_portalFrustum, float p_elapsedTimef)
{
	COrient3D vp;

	// First, set up for pseudo-ortho
	vp = p_render->GetViewpoint();

	p_render->PushMatrix();
	p_render->LoadIdentity();
	p_render->Rotate(180,0,1,0);
	p_render->Translatef(0,0,1);

	// And ignore depth testing and writing
	p_render->EnableDepthTest(false);

	CPtlDoor *collisionDoor;
	CSurface3D *collisionSurface;
	CPtlRoom *collisionRoom;
	CVector3D collisionPoint;
	float distance;
	BYTE alpha;
	float xo, yo, zo;

	// Loop through the light sources
	CVector3D start = p_portalFrustum.GetViewpoint();
	CLightSourceListNode *node = lightSources.GetFirst();
	if (node)
	while (node->next != NULL)
	{
		if (node->visible)
		{
			// Is point inside frustum?
			//if (p_portalFrustum.Point

			// Check for wall collision (note: alpha surfaces will work too, 
			//   alpha won't be visible through them
			CVector3D direction = node->position - start;
			if (direction.Unitize())
			{
				// Just need to check from viewpoint to light
				distance = (node->position - start) * direction;
				node->flareRadius = 0.05f * node->radius / float(pow(distance, 1.2f)) * node->intensity;

				if (RayCollide(start, direction, distance, collisionPoint, &collisionSurface, &collisionRoom, &collisionDoor))
					node->targetFlareSize = 0.0f;
				else
					node->targetFlareSize = 1.0f;

			//	node->currentFlareRadius = 0.0f;

				if (node->targetFlareSize != node->currentFlareSize)
				{
					if (node->currentFlareSize < node->targetFlareSize )
					{
						node->currentFlareSize += (0.02f * p_elapsedTimef);

						if (node->currentFlareSize > node->targetFlareSize)
							node->currentFlareSize = node->targetFlareSize;
					}
					else
					{
						node->currentFlareSize -= (0.02f * p_elapsedTimef);

						if (node->currentFlareSize < node->targetFlareSize)
							node->currentFlareSize = node->targetFlareSize;
					}
				}

				// Render the alpha flare
				if (node->currentFlareSize > 0.0f)
				{
					if (distance < 1.0f)
					{
						alpha = 240;
					}
					else
					{
						//alpha = BYTE(240.0f / pow(distance, 0.7f));
						alpha = BYTE(240.0f / pow(distance, 0.2f));
					}
					xo = vp.l*(node->position - start);
					yo = vp.u*(node->position - start);
					zo = vp.f*(node->position - start);
					if (zo > 0.001f)
					{
						xo = xo/zo;
						yo = yo/zo;

						{
							p_render->PushMatrix();
							p_render->Translatef(xo, yo, 0.0f);
							//TRACE("Rendering flare at %g * %g = %g\n", node->currentFlareSize, node->flareRadius, node->currentFlareSize * node->flareRadius);
							p_render->Scalef(node->currentFlareSize * node->flareRadius, node->currentFlareSize * node->flareRadius, 1.0f);
							p_render->RenderQuad(node->color, alpha, APITextures.GetTexture("hh1"));
							p_render->PopMatrix();
						}
					} // zo > 0.001
				}
			}	// direction.Unitize
		}

		node = (CLightSourceListNode *)(node->next);
	}

	// Restore API viewpoint and depth state
	p_render->EnableDepthTest(true);
	p_render->PopMatrix();
}

// Determine which room the point is in, return -1 if none
// Uses the octree and intList of the cells
int CPtlMap::RoomOfPoint(CVector3D &p_point)
{
	COctree *cell = octree.PointIn(p_point);
	if (cell)
	{
		CLinkedListNode<CInteger> *inode = cell->intList.GetFirst();

		if (inode)
		while (inode->next != NULL)
		{
			if (rooms[inode->value].PointInRoom(p_point, vertices))
				return inode->value;

			inode = inode->next;
		}
	}

	return -1;
}

void CPtlMap::PlaceParticleEmitter(CLinkedListNode<CParticleEmitter> *p_particleEmitter)
{
	CLinkedListNode<CParticleEmitter> *node = p_particleEmitter;
	if (!node) return;

	COctree *cell;

	// Move it to a new room if need be
	if ((node->oldPosition != node->position) || (node->intRef == -1))
	{
		// We need old position to properly calculate positions of new particles
		//    When we animate the emitter, this will be set
		//node->oldPosition = node->position;

		// Get the octree cell this particle is in
		// (This code is mostly identical to the code below for particles,
		//   except it's for the emitter)
		cell = octree.PointIn(node->position);

		if (cell)
		{
			// Get all the rooms the cell touches
			// Find the room the emitter is in
			CLinkedListNode<CInteger> *inode = cell->intList.GetFirst();
			if (inode)
			{
				// If there is only one room in the list, emitter
				//   belongs there
				if (inode->next->next == NULL)
				{
					// Only reassign if it's a different room than this one
					// (id is a class member)
					if (inode->value != node->intRef)
					{
						if (node->intRef != -1)
							node->Detach();

						rooms[inode->value].AddDetachedParticleEmitter(node);
						node->intRef = inode->value;
					}
				}
				else
				{
					// There is more than one room in the list, figure
					//   out which room we are in
					// First make sure we have moved out of our current room
					bool done = false;
					if (node->intRef != -1)
						if (rooms[node->intRef].PointInRoom(node->position, vertices))
							done = true;

					if (!done)
					{
						bool found = false;
						// See if the particle passed through a portal in the current
						//   room - if it has, assign the particle to the room on the
						//   other side
						// ??

						if (!found)
						{
							// Now figure out which room we are in the long way
							while ((inode->next != NULL) && (!found))
							{
								if (inode->value != node->intRef)
								if (rooms[inode->value].PointInRoom(node->position, vertices))
								{
									found = true;

									if (node->intRef != -1)
										node->Detach();

									rooms[inode->value].AddDetachedParticleEmitter(node);
									node->intRef = inode->value;
								}

								inode = inode->next;
							}
						}
					} // point not in assigned room anymore
				} // more than one room in the list
			}	// room list exists
		}	// found a cell list
	}
}
