// PtlMap.h: interface for the CPtlMap class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PTLMAP_H__7389B605_62F5_4441_8DC6_F2CF992CD613__INCLUDED_)
#define AFX_PTLMAP_H__7389B605_62F5_4441_8DC6_F2CF992CD613__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "PtlPortal.h"
#include "PtlRoom.h"
#include "..\api3d\ModelVertex3D.h"
#include "..\api3d\RenderObject.h"
#include "..\structs\octree.h"

#define MAXMIRRORDEPTH  8
#define REFLECTEDVPADJUST 0.001f

typedef CLinkedListNode<CInteger> * LPINTNODE;
#define MAXMIRRORDEPTH 8

class CPtlMap  
{
public:
	CPtlMap();
	virtual ~CPtlMap();
	void Destroy(void);

	static unsigned int renderID;

private:
	CPtlPortal *portals;
	unsigned portalQty;

	CPtlRoom *rooms;
	unsigned roomQty;

	CModelVertex3D *vertices;
	unsigned vertexQty;

	unsigned mirrorDepth;	// Incremented/Decremented during rendering, 
							//   depth of mirrors we are rendering inside

	// for mirror clip planes
	CVector3D mirrorPoints[MAXMIRRORDEPTH];
	CVector3D mirrorNormals[MAXMIRRORDEPTH];

	// This will be moved to the room level eventually (actually, leaving
	//   them here for now since only references in octree allow them
	//   to be used in rendering)
	CLightSourceList lightSources;

	// Objects to render (must have references in octree to be used)
	CRenderObjectList renderObjects;

	CLinkedList<CLPVoid> vpList;	// Void pointer list (temp for now)

	COctree	octree;	// Octree for the map (holds references to objects,
					//   lightsources and particle sets)
	CVector3D minPoint;		// Bounding volume for entire map
	CVector3D maxPoint;

public:
	void Initialize(unsigned p_vertexQty, unsigned p_roomQty, unsigned p_portalQty);

	void SetVertex(unsigned p_id, float p_x, float p_y, float p_z)
	{
		vertices[p_id].vertex.SetVec(p_x, p_y, p_z);
	}

	CPtlRoom & GetRoom(unsigned p_id) {return rooms[p_id];}
	unsigned GetRoomQty(void) {return roomQty;}

	CPtlPortal & GetPortal(unsigned p_id) {return portals[p_id];}

	CModelVertex3D * GetVertices(void) {return vertices;}
	CModelVertex3D & GetVertex(unsigned p_id) {return vertices[p_id];}

	inline void CalculateSurfaceData(void)
	{
		CalculateSurfaceData(1000.0f, 0.0f);
	}
	inline void CalculateSurfaceData(float p_targetVolume)
	{
		CalculateSurfaceData(p_targetVolume, 0.0f);
	}
	void CalculateSurfaceData(float p_targetVolume, float p_minDimension);

	void SetTwoOpposingSurfaces(unsigned p_roomID, unsigned p_surfaceID, unsigned p_opposingRoomID, unsigned p_opposingSurfaceID);

	// This won't be necessary when lightsources are moved to the room level
	CLightSourceList * GetLightSourceList(void) {return &lightSources;}

	inline COctree & GetOctree(void) {return octree;}

	int RoomOfPoint(CVector3D &p_point);

	//////////////////////////////////////////////
	// Rendering routines

	// Main render routine - call this from the main program, and to
	//   start a subrender
	void StartRender(IRender *p_render, CPtlFrustum &p_portalFrustum, DWORD p_elapsedTimeMS, DWORD p_currentTime)
	{
		renderID++;
		Render(p_render, p_portalFrustum, p_elapsedTimeMS, p_currentTime, -1, true);
	}
	void StartRender(IRender *p_render, CPtlFrustum &p_portalFrustum, DWORD p_elapsedTimeMS, DWORD p_currentTime, int p_startingRoomID)
	{
		renderID++;
		Render(p_render, p_portalFrustum, p_elapsedTimeMS, p_currentTime, p_startingRoomID, true);
	}
	void StartRender(IRender *p_render, CPtlFrustum &p_portalFrustum, DWORD p_elapsedTimeMS, DWORD p_currentTime, int p_startingRoomID, bool p_postparseRender)
	{
		renderID++;
		Render(p_render, p_portalFrustum, p_elapsedTimeMS, p_currentTime, p_startingRoomID, p_postparseRender);
	}

private:
	void Render(IRender *p_render, CPtlFrustum &p_portalFrustum, DWORD p_elapsedTimeMS, DWORD p_currentTime, int p_startingRoomID, bool p_postparseRender);

public:
	void RenderFloorTexture(IRender *p_render, CVector3D &p_pos, float p_radius, float p_rotation, CAPITexture *p_texture, CColorRGB *p_modColor, BYTE p_alpha);

	void RenderLightFlares(IRender *p_render, CPtlFrustum &p_portalFrustum, float p_elapsedTimef);

private:
	void RenderRoom(int p_roomID, IRender *p_render, CPtlFrustum &p_portalFrustum, bool p_postparseRender, LPINTNODE *p_roomsRendered, LPINTNODE p_lastNode, int p_lastRoomID, DWORD p_elapsedTimeMS, DWORD p_currentTime);

	// These will move eventually
	CRenderObject controlObject;
	CRenderObject testObject;

	CColorRGB	ambientColor;	// Used to calculate simulated lighting
								//    on objects
public:
	void SetAmbientColor(BYTE p_red, BYTE p_green, BYTE p_blue)
	{
		ambientColor.SetColor(p_red, p_green, p_blue);
	}
	void SetAmbientColor(CColorRGB p_color)
	{
		ambientColor.SetColor(p_color);
	}
	void AnimateDoors(long p_elapsedTimeMS);

	bool RayCollide(CVector3D &p_start, CVector3D &p_directionUnit, float &p_distanceOut, CVector3D &p_collisionPointOut, CSurface3D **p_collisionSurfaceOut, CPtlRoom **p_collisionRoomOut, CPtlDoor **p_collisionDoorOut);
	bool SphereCollision(CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CSurface3D **p_collisionSurfaceOut, CVector3D &p_collisionWorldNormalOut, CPtlRoom **p_collisionRoomOut, CPtlDoor **p_collisionDoorOut);

	CRenderObject * AddRenderObject(void)
	{
		return renderObjects.NewNode();
	}

	// These will eventually go away
	CLightSourceListNode * GetLightSource1(void)
	{
		return lightSources.GetFirst();
	}
	CLightSourceListNode * GetLightSource2(void)
	{
		if (lightSources.GetFirst())
		{
			if (lightSources.GetFirst()->next->next != NULL)
				return (CLightSourceListNode *)(lightSources.GetFirst()->next);
			else
				return NULL;
		}
		else
			return NULL;
	}
	CRenderObject * GetControlObject(void) {return &controlObject;}
	CRenderObject * GetTestObject(void) {return &testObject;}

	void * GetVoidPtr1(void)
	{
		CLinkedListNode<CLPVoid> *node = vpList.GetFirst();
		if (node)
			return (node->lpVoid);
		else
			return NULL;
	}
	CLinkedListNode<CLPVoid> * AddVP(void)
	{
		return vpList.NewNode();
	}

	void PlaceParticleEmitter(CLinkedListNode<CParticleEmitter> *p_particleEmitter);
};

#endif // !defined(AFX_PTLMAP_H__7389B605_62F5_4441_8DC6_F2CF992CD613__INCLUDED_)
