// PtlPortal.cpp: implementation of the CPtlPortal class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PtlPortal.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPtlPortal::CPtlPortal()
{
	frontRoomID = -1;
	backRoomID = -1;

	mirror = false;
	closed = false;

	mirrorSurface = NULL;

	vertexRefs = NULL;
	vertexRefQty = 0;

	normal.SetVec(0,0,0);
}

CPtlPortal::~CPtlPortal()
{
	Destroy();
}

void CPtlPortal::Destroy(void)
{
	// Get rid of doors
	doorList.EmptyList();

	if (vertexRefs)
		delete [] vertexRefs;
	vertexRefs = NULL;
	vertexRefQty = 0;
}

void CPtlPortal::Initialize(unsigned p_vertexRefQty)
{
	vertexRefQty = p_vertexRefQty;
	if (vertexRefQty)
		vertexRefs = new int[vertexRefQty];

	// Just in case
	normal.SetVec(0,0,0);
}

bool CPtlPortal::IsClosed(void)
{
	// If portal set closed overrides all
	if (closed) return true;

	// If all doors are closed, the portal is closed
	if (!doorList.IsEmpty())
	{
		CLinkedListNode<CPtlDoor> *node;
		node = doorList.GetFirst();
		if (node)
		while (node->next != NULL)
		{
			if (!node->IsClosed())
				return false;

			node = node->next;
		}
	}
	else
		return false;

	// Al doors are closed - portal is closed
	return true;

}

void CPtlPortal::CalculateData(CModelVertex3D *p_vertices)
{
    static CVector3D v1;
    static CVector3D v2;
    
	// Ortho version (only x and y set)
    if (vertexRefQty == 2)
	{
        if ((p_vertices[vertexRefs[0]].vertex.z == 0) && (p_vertices[vertexRefs[1]].vertex.z == 0))
		{
            v1 = p_vertices[vertexRefs[1]].vertex - p_vertices[vertexRefs[0]].vertex;
            normal.SetVec(-v1.y, v1.x, 0.0f);
            normal.Unitize();
		}
        else
            normal.SetVec(0, 0, 0);
    }
        
    if (vertexRefQty > 2)
	{
        // Make sure we get a good one, can't make a normal off
        //   parallel segments
        unsigned i;
		normal.SetVec(0,0,0);
        for (i=0; i<=vertexRefQty - 3; i++)
		{
            v1 = p_vertices[vertexRefs[i+0]].vertex - p_vertices[vertexRefs[i+1]].vertex;
            v2 = p_vertices[vertexRefs[i+2]].vertex - p_vertices[vertexRefs[i+1]].vertex;
            normal = v1.CrsProd(v2);
            if (normal.MgnTudSqr() > 0.00001f)
			{
                normal.Unitize();
                i = vertexRefQty - 3;
			}
        }

        if (normal.MgnTudSqr() == 0)
            TRACE("** Warning: Zero portal normal on PtlPortal object");
    }
}

// Is the portal visible given the frustum?
bool CPtlPortal::IsVisible(CPtlFrustum &p_frustum, CModelVertex3D *p_vertices, int p_roomID)
{
	static CVector3D v1;
	unsigned i;

	// First check to see if the portal faces the viewpoint
	//   (its corresponding side in a room should be facing the viewpoint)
	// If it doesn't, it isn't visible
	// We allow dp = 0 when we are right on a portal and can see both rooms
	//   (standing in a doorway facing sideways), but this requires
	//   a special check to avoid an infinite loop (see RenderRoom)
    bool ok = false;
    if (frontRoomID == p_roomID)
	{
        float dp = (p_frustum.GetViewpoint() - p_vertices[vertexRefs[0]].vertex) * normal;
        if ((dp > 0) || ((dp > -0.0001f) && (!mirror)))
            ok = true;
	}
	else
	{
        if (backRoomID == p_roomID)
		{
			float dp = (p_frustum.GetViewpoint() - p_vertices[vertexRefs[0]].vertex) * normal;
			if ((dp < 0) || ((dp < 0.0001f) && (!mirror)))
				ok = true;
		}
	}

    if (!ok)
		return false;

    // This loop will quickly fail on points that aren't
    //   inside, and will end with the first one that is inside all
    //   It's faster overall than looking for one plane for which
    //   all are outside to invalidate the whole polygon
    i=0;
	bool inside = false;
	CLinkedListNode<CVector3D> *node;
    while ((i <= vertexRefQty - 1) && (!inside))
	{
        inside = true;
        v1 = p_vertices[vertexRefs[i]].vertex - p_frustum.GetViewpoint();
		node = p_frustum.GetPlaneNormalList().GetFirst();
		if (node)
		{
			while ((node->next != NULL) && (inside))
			{
				if ((v1 * (*node)) < 0.0f)
					inside = false;

				node = node->next;
			}
			i++;
		}
		else
			// We have no normals, how could it be visible?
			return false;
    }
    
    // If there is a point inside all the normals, it's visible
    if (inside)
		return true;

	if (vertexRefQty > 2)
	{
		// Splice portal with frustum
		// If we get a result, it's visible.
		// This is a test only, it's expensive.
		// It's more accurate, though
		// Only do this with 3D
	//    If splicetest Then
		static CVector3D verts[64];
		static CVector3D vertsOut[64];
		CVector3D *a1, *a2, *aTemp;
		unsigned a1Qty, a2Qty;
		//unsigned aQTemp;;

		a1 = verts; a2 = vertsOut;

		static unsigned vertexQty;
		// ?? CopyMemory would be faster here
		//   alas, the data isn't sequential in memory
		for (i=0; i<vertexRefQty; i++)
			verts[i] = p_vertices[vertexRefs[i]].vertex;
		a1Qty = vertexRefQty;

		CLinkedListNode<CVector3D> *node;
		node = p_frustum.GetPlaneNormalList().GetFirst();
		if (node)
		{
			while (node->next != NULL)
			{
				if (!SplicePolygonWithPlane(a1, a1Qty, a2, a2Qty, p_frustum.GetViewpoint(), *node))
					return false;
				else
				{
					// Swap the pointers and quantities for input/output
					aTemp = a1;
					//aQtemp = a1Qty;

					a1 = a2;
					a1Qty = a2Qty;

					a2 = aTemp;
					//a2Qty = aQTemp;
				}

				node = node->next;
			}
		}
		else
			// we have no normals, how could it be visible?
			return false;


		// We got a non-null splice, the portal is visible
		return true;
	}
	else
		if (vertexRefQty == 2)
		{
			// This is for the Ortho version of portals
			// Check line segment intersections
			// ??
			return false;
		}

	// Leftover
	return false;

}

bool CPtlPortal::MergeFrustums(CPtlFrustum &p_oldFrustum, CPtlFrustum &p_newFrustum, CModelVertex3D *p_vertices, int p_currentRoomID)
{
	// We could just put all the normals from the old frustum and the normals
	//   made from the portal into one frustum and get a valid one,
	//   but we should try to cut out the superfluous ones

	// This routine will be somewhat forgiving against floating point truncation,
	//   however, because we would rather have too many planes than too few

	// If return false, we got a null result

	// ?? Throughout this routine, some vector variables hold what we need
	//    but aren't taking advantage of them
	//    and others could be put into variables to speed it up

	unsigned normalQty;	// if static, can't debug
	unsigned i, j;
	static CVector3D v1, v2, v3;
	float dp, t;
	bool keep;
	CVector3D vp;
	CPlaneNormalListNode *node, *node1;

	//TRACE("***** Begin MergeFrustums\n");

	normalQty = 0;

	// New frustum will use same viewpoint and forward vector
	p_newFrustum.SetViewpoint(p_oldFrustum.GetViewpoint(), p_oldFrustum.GetForward());
	vp = p_newFrustum.GetViewpoint();

	// Keep any planar normals from the old frustum for which and portal points
	//   are outside.  Discard any for which all points are inside.
//	char out[256];
	node = p_oldFrustum.GetPlaneNormalList().GetFirst();
	if (node)
	while (node->next != NULL)
	{
		keep = false;
		//TRACE ("Normal being tested: %s\n", node->OutStr(out));
		for (j=0; j<vertexRefQty; j++)
		{
			v1 = p_vertices[vertexRefs[j]].vertex - vp;
			//TRACE ("Values: point: %s\n", p_vertices[vertexRefs[j]].vertex.OutStr(out));
			//TRACE ("Values: viewpoint: %s\n", p_newFrustum.GetViewpoint().OutStr(out));
			//TRACE ("Offset being tested: %s\n", v1.OutStr(out));

			if ((v1 * (*node)) < 0.0f)
			{
				keep = true;
				j = vertexRefQty - 1;
			}
		}

		if (keep)
		{
			//TRACE ("Normal kept\n");
			p_newFrustum.AddPlaneNormal(*node);
			normalQty = normalQty++;
		}

		node = (CPlaneNormalListNode *)(node->next);
	}

	// Now form planar normals out of portal segments that are entirely inside or
	//   visible cut into the old planes we kept (current new frustum)
	if (vertexRefQty == 2)
	{
		// Ortho solution
		// ??
	}
	else
		if (vertexRefQty > 2)
		{
			// Which side of the portal are we on?
			//   (Affects how to build a proper normal out of a portal segment)
			bool front;
//			if (((vp - p_vertices[vertexRefs[0]].vertex) * normal) > 0)
			if (p_currentRoomID == frontRoomID)
				front = true;
			else
				front = false;

			// v1 is the segment, v3 is the starting point
			// v2 will be a segment intersection with a plane
            v1 = p_vertices[vertexRefs[0]].vertex - p_vertices[vertexRefs[vertexRefQty - 1]].vertex;
            v3 = p_vertices[vertexRefs[vertexRefQty - 1]].vertex;
			// Slightly different loop here, 1 to vertexRefQty instead of 0 to qty-1
			for (i=1; i<=vertexRefQty; i++)
			{
				//TRACE("Segment being tested: %s\n", v1.OutStr(out));
				//TRACE("Point being tested: %s\n", v3.OutStr(out));

				// Test the starting point, if it's inside we use this segment
				keep = true;

				CPlaneNormalListNode *node;
				node = p_oldFrustum.GetPlaneNormalList().GetFirst();
				if (node)
				while ((node->next != NULL) && (keep))
				{
					if (((v3 - vp) * (*node)) < 0.0f)
					{
						keep = false;
					}

					node = (CPlaneNormalListNode *)(node->next);
				}

				if (keep)
				{
					//TRACE("Kept it 1");
					// Use this segment, form a normal
					if (front)
                        p_newFrustum.AddPlaneNormal((p_vertices[vertexRefs[i - 1]].vertex - vp).CrsProd(v1));
					else
                        p_newFrustum.AddPlaneNormal((vp - p_vertices[vertexRefs[i - 1]].vertex).CrsProd(v1));

					normalQty = normalQty++;
				}
				else
				{
					// Still don't know if we want to keep it -
					//  See if the segment cuts visibly into the frustum
					// Get any intersection point with a normal, and if the
					//    collision point lies within all the other normals,
					//    it cuts visibly
					node = p_oldFrustum.GetPlaneNormalList().GetFirst();
					if (node)
					while (node->next != NULL)
					{
						dp = v1 * (*node);
						//TRACE("Testing with normal: %s\n", node->OutStr(out));
						if (fabs(dp) > 0.001f)
						{
							// NOTE: We could be testing this against the
							//   new frustum list instead of the old frustum
							//   list, but this helps guarantee we have enough normals,
							//   if not too many
							t = -(*node * (v3 - vp)) / dp;
							if ((t >= -0.001f) && (t <= 1.001f))
							{
								v2 = v3 + (v1 * t);
								//TRACE(">>Intersection: %s\n", v2.OutStr(out));

								// v2 is the intersection point, node is
								//   the normal we intersected
								// See if v2 is visible to all the other normals
								keep = true;
								node1 = p_oldFrustum.GetPlaneNormalList().GetFirst();
								if (node1)
								while ((node1->next != NULL) && (keep))
								{
									if (node1 != node)
									{
										if ((((v2 - vp) * (*node1)) ) < -0.001f)
											keep = false;
									}

									node1 = (CPlaneNormalListNode *)(node1->next);
								}

								if (keep)
								{
									//TRACE("Kept normal\n");
									if (front)
										p_newFrustum.AddPlaneNormal((p_vertices[vertexRefs[i - 1]].vertex - vp).CrsProd(v1));
									else
										p_newFrustum.AddPlaneNormal((vp - p_vertices[vertexRefs[i - 1]].vertex).CrsProd(v1));

									normalQty++;
								}

							}	// t in range
						}	// fabs(dp) > 0.001

						node = (CPlaneNormalListNode *)(node->next);
					}	// while node

				}	// test if cuts visibly into frustum

				// Get next segment and starting point to test
                if (i < vertexRefQty)
				{
                    v1 = p_vertices[vertexRefs[i]].vertex - p_vertices[vertexRefs[i - 1]].vertex;
                    v3 = p_vertices[vertexRefs[i - 1]].vertex;
                }

			} // for i = 1 to vertexRefQty
		
		}	// vertexRefQty > 2

	/*TRACE("***** Done MergeFrustums\n");

	TRACE("Resulting normal list:\n");
	node = p_newFrustum.GetPlaneNormalList().GetFirst();
	if (node)
	while (node->next != NULL)
	{
		TRACE("%s\n", node->OutStr(out));
		node = (CPlaneNormalListNode *)node->next;
	}*/

	if (normalQty > 0)
		return true;
	else
		return false;
}

bool CPtlPortal::SplicePolygonWithPlane(CVector3D *p_vertices, unsigned p_vertexQty, CVector3D *p_verticesOut, unsigned &p_vertexQtyOut, CVector3D &p_planePoint, CVector3D &p_planeNormal)
{
	// If zero vertices are the result, return false

	p_vertexQtyOut = 0;

	bool firstPointInFront;		// was first point in front?
	bool lastPointInFront;		// was previous point in front?

	static CVector3D v1, v2;
	static float dp;
	static float k;
	static CVector3D pt;

	unsigned i;
	for (i=0; i<p_vertexQty; i++)
	{
		if (((p_vertices[i] - p_planePoint) * (p_planeNormal)) >= -0.0001f)
		{
			if (i == 0)
				firstPointInFront = true;
			else
			{
				if (!lastPointInFront)
				{
					// Get entry point
					v1 = p_vertices[i] - p_vertices[i - 1];
					v2 = p_vertices[i - 1] - p_planePoint;
					dp = p_planeNormal * v1;
					if (fabs(dp) > 0.00001f)
					{
						k = -(p_planeNormal * v2) / dp;
						if ((k >= -0.001f) && (k <= 1.001f))
						{
							// Might need an adjustment
							// ??
							// pt.SetVec2(p_vertices(i - 1).AddVec(v1.ScmVec(k - 0.001 / v1.MgnTud)))
							pt = p_vertices[i - 1] + (v1 * k);
						}
						else
						{
							TRACE ("** Failed splice");
							p_vertexQtyOut = 0;
							return false;
						}
					}
					else
					{
						TRACE ("** Failed splice");
						p_vertexQtyOut = 0;
						return false;
					}
                
					// Add entry point
					p_verticesOut[p_vertexQtyOut] = pt;
					p_vertexQtyOut++;
				}
			}
        
			// Add point
			p_verticesOut[p_vertexQtyOut] = p_vertices[i];
			p_vertexQtyOut++;
        
			lastPointInFront = true;
		}
		else
		{
			if (i == 0)
				firstPointInFront = false;
			else
			{
				if (lastPointInFront)
				{
					// Get exit point
					v1 = p_vertices[i] - p_vertices[i - 1];
					v2 = p_vertices[i - 1] - p_planePoint;
					dp = p_planeNormal * v1;
					if (fabs(dp) > 0.00001f)
					{
						k = -(p_planeNormal * v2) / dp;
						if ((k >= -0.001f) && (k <= 1.001f))
						{
							// Might need an adjustment
							// ??
							//Call pt.SetVec2(p_vertices(i - 1).AddVec(v1.ScmVec(k - 0.001 / v1.MgnTud)))
							pt = p_vertices[i - 1] + (v1 * k);
						}
						else
						{
							TRACE ("** Failed splice");
							p_vertexQtyOut = 0;
							return false;
						}
					}
					else
					{
						TRACE("** Failed splice");
						p_vertexQtyOut = 0;
						return false;
					}
                
					// Add exit point
					p_verticesOut[p_vertexQtyOut] = pt;
					p_vertexQtyOut++;
				}
			}
        
			lastPointInFront = false;
		}
	}
    
	// Take care of final segment
	if ((!lastPointInFront) && (firstPointInFront))
	{
		// Get entry point between vertex n-1 and 0
		v1 = p_vertices[0] - p_vertices[p_vertexQty-1];
		v2 = p_vertices[p_vertexQty-1] - p_planePoint;
		dp = p_planeNormal * v1;
		if ( fabs(dp) > 0.00001f )
		{
			k = -(p_planeNormal *v2) / dp;
			if ((k >= -0.001f) && (k <= 1.001f))
			{
				// Might need an adjustment
				// ??
				//Call pt.SetVec2(p_vertices(UBound(p_vertices)).AddVec(v1.ScmVec(k - 0.001 / v1.MgnTud)))
				pt = p_vertices[p_vertexQty-1] + (v1 * k);
			}
			else
			{
				TRACE ("** Failed splice");
				p_vertexQtyOut = 0;
				return false;
			}
		}
		else
		{
			TRACE ("** Failed splice");
			p_vertexQtyOut = 0;
			return false;
		}

            
		// Add entry point
		p_verticesOut[p_vertexQtyOut] = pt;
		p_vertexQtyOut++;
	}
	if ((lastPointInFront) && (!firstPointInFront))
	{
		// Get exit point between vertex n-1 and 0
		v1 = p_vertices[0] - p_vertices[p_vertexQty-1];
		v2 = p_vertices[p_vertexQty-1] - p_planePoint;
		dp = p_planeNormal * v1;
		if (fabs(dp) > 0.00001f)
		{
			k = -(p_planeNormal * v2) / dp;
			if ((k >= -0.001f) && (k <= 1.001f))
			{
				// Might need an adjustment
				// ??
				//Call pt.SetVec2(p_vertices(UBound(p_vertices)).AddVec(v1.ScmVec(k + 0.001 / v1.MgnTud)))
				pt = p_vertices[p_vertexQty-1] + (v1 * k);
			}
			else
			{
				TRACE ("** Failed splice");
				p_vertexQtyOut = 0;
				return false;
			}
		}
		else
		{
			TRACE ("** Failed splice");
			p_vertexQtyOut = 0;
			return false;
		}
    
		// Add exit point
		p_verticesOut[p_vertexQtyOut] = pt;
		p_vertexQtyOut++;
	}

/*leave:

' Output this test case for debugging
Open "failedsplice.txt" For Append As #1
Print #1, "Plane point"
Print #1, p_planePoint.OutputString
Print #1, "Plane normal"
Print #1, p_planeNormal.OutputString
Print #1, "Vertices"
For i = 0 To UBound(p_vertices)
    Print #1, p_vertices(i).OutputString
Next i
Print #1, "dp= " & dp
Print #1, "k= " & k
Print #1, "------------"
Close #1

ReDim result(0)
SplicePolygonWithPlane = result

leave2:*/

	if (p_vertexQtyOut == 0)
		return false;
	else
		return true;
}

void CPtlPortal::RenderSpliced(IRender *p_render, CModelVertex3D *p_vertices, CPtlFrustum &p_frustum, CVector3D *p_verticesOut, unsigned &p_vertexQtyOut)
{
	// If this routine returns a p_vertexQtyOut of zero, it failed

	// Splice the portal surfec to the current frustum and render the
	//   result

    // Also return the resulting vertex array
    
    static CVector3D vert1[64];
	static CVector3D vert2[64];
	static CVector3D *a1, *a2, *aTemp;
	static unsigned a1Qty, a2Qty;
    
    // First, splice the portal to the frustum
    // Build an initial array of points
	// CopyMemory would be nice, but the data isn't sequential in memory
    unsigned i;
    for (i=0; i<vertexRefQty; i++)
        vert1[i] = p_vertices[vertexRefs[i]].vertex;

    // Splice it
    // We are passiing the polygon to each plane, getting
    //   a result, and pssing it to the next plane.
    // It's slow doing it this way, but at least we
    //   get a correct result until I can come up with a better way
    // (When I know I don't have any superfluous planes, and
    //   they are all adjacent in order)
	CPlaneNormalListNode *node;
	a1 = vert1; a2 = vert2; a1Qty = vertexRefQty;
	node = p_frustum.GetPlaneNormalList().GetFirst();
	if (node)
	while (node->next != NULL)
	{
		if (!SplicePolygonWithPlane(a1, a1Qty, a2, a2Qty, p_frustum.GetViewpoint(), *node))
		{
			p_vertexQtyOut = 0;
			return;
		}
		else
		{
			// Swap the inputs and outputs
			aTemp = a1;

			a1 = a2;
			a1Qty = a2Qty;
			//TRACE("# vertices: %i\n", a2Qty);

			a2 = aTemp;
		}

		node = (CPlaneNormalListNode *)node->next;
	}
	// Provide result to calling routine
	p_vertexQtyOut = a1Qty;
	for (i=0; i<p_vertexQtyOut; i++)
	{
		p_verticesOut[i] = a1[i];
	}
	//CopyMemory(p_verticesOut, a1, a1Qty*12);
	//TRACE("Number of vertices in mirror surface: %i\n", p_vertexQtyOut);

    // If it's valid, draw it
    CSurface3D surface;		// If you static this, you'll get a memory leak
    CModelVertex3D vertices[64];
    
    surface.Initialize(p_vertexQtyOut);
    for (i=0; i<p_vertexQtyOut; i++)
	{
        vertices[i].vertex = p_verticesOut[i];
        surface.SetVertexRef(i, i);
    }

    // Render it and return
	// This might be better is we have a surace with 64 vertices and somehow
	//   tell it to only render a certain number of vertices
	// That way we aren't destroying a surface with its vertexRefs all the
	//   time
	// ??
    p_render->RenderSurfaces(&surface, 1, vertices, NULL, 255);
}

void CPtlPortal::RenderVertexArray(IRender *p_render, CVector3D *p_vertexArray, unsigned p_vertexQty, bool p_renderSurface, CAPITexture **p_lightmaps, CLightSourceRefList *p_lightSourceList)
{
    // Assemble a stored spliced surface and render it
    // The vertices here were stored from a previous splice,
    //   it doesn't really use the vertices of CPtlMap
    
    // if p_renderSurface is true, we'll render a visible surface
    //   if one is available.  Otherwise we aren't writing color
    
    CSurface3D surface;	// If you static this, you'll get a memory leak
    CModelVertex3D vertices[64];
    
    bool renderSurface = false;
    if (p_vertexQty > 0)
	{
		if (!p_renderSurface)
            renderSurface = false;
        else
		{
            if (!mirrorSurface)
                renderSurface = false;
            else
                renderSurface = mirrorSurface->IsVisible();
		}
        
        if (!renderSurface)
		{
            p_render->EnableColorWrite(false);
            
            surface.Initialize(p_vertexQty);
            unsigned i;
            for (i=0; i<p_vertexQty; i++)
			{
                vertices[i].vertex = p_vertexArray[i];
                surface.SetVertexRef(i, i);
			}
            
            p_render->RenderSurfaces(&surface, 1, vertices, NULL, 255);
            
            p_render->EnableColorWrite(true);
        }
		else
		{
            // Render the surface (assumed alpha, but we'll use the surface attributes anyway)
            // Assemble a surface using our vertex array but decal it
            //  across the mirror surface (calculate tex coords, etc)
            surface.Initialize(p_vertexQty, mirrorSurface->GetTexture());
			float s, t;
			unsigned i;
            for (i=0; i<p_vertexQty; i++)
			{
                vertices[i].vertex = p_vertexArray[i];
                // Need color, would have to interpolate.  Assume white for now
                // ??
                vertices[i].color.SetColor(255, 255, 255);
                if (surface.GetTexture())
				{
                    mirrorSurface->CalculateTex1Coords(p_vertexArray[i], s, t);
                    surface.SetVertexRefTexture(i, i, s, t);
                    // If it has a light map, we'll have to calculate those too
                    // ??
                    // And decals!  oh boy.  Maybe we can render those in the
                    //   main loop?
                    // We could take advantage of the stencil before we erase it...
				}
                else
				{
                    surface.SetVertexRef(i, i);
				}
                surface.SetLightmapID(mirrorSurface->GetLightmapID());
                if (surface.GetLightmapID() != -1)
				{
                    mirrorSurface->CalculateTexLMCoords(p_vertexArray[i], s, t);
                    surface.SetVertexLMCoords(i, s, t);
				}
            }
            surface.SetAlpha(mirrorSurface->GetAlpha());
            // We need the normal for applying light sources
            surface.SetNormal(mirrorSurface->GetNormal());
            // And the values to calculate LM coords for applying dynamic light sources
            surface.SetLMOrigin(mirrorSurface->GetLMOrigin());
            surface.SetLMAxisS(mirrorSurface->GetLMAxisS());
            surface.SetLMAxisT(mirrorSurface->GetLMAxisT());
			// And the coordinates assigned to it for the offscreen buffer
			surface.SetDynLMOrigin(mirrorSurface->GetDynLMX(), mirrorSurface->GetDynLMY());
            
            if (mirrorSurface->UseSurfaceColor())
                surface.SetColor(mirrorSurface->GetColor());
            
            p_render->RenderSurfaces(&surface, 1, vertices, NULL, 255, p_lightmaps, NULL, false, p_lightSourceList);
        }
    }
    
}

/////////////////////////////////////////////////////////////
// Door routines

void CPtlPortal::RenderDoors(IRender *p_render, CLightSourceList *p_lightSourceList, unsigned int p_renderID)
{
    static CPtlDoorListNode *node;
    
    if (!doorList.IsEmpty())
	{
        node = (CPtlDoorListNode *)doorList.GetFirst();
		if (node)
        while (node->next != NULL)
		{
			if (node->renderID != p_renderID)
			{
				node->renderID = p_renderID;
				node->Render(p_render);
			}
            
            node = (CPtlDoorListNode *)node->next;
        }
    }
}

void CPtlPortal::AnimateDoors(long p_elapsedTimeMS)
{
    static CPtlDoorListNode *node;
    
    if (!doorList.IsEmpty())
	{
        node = (CPtlDoorListNode *)doorList.GetFirst();

		if (node)
        while (node->next != NULL)
		{
            node->Animate (p_elapsedTimeMS);
            
            node = (CPtlDoorListNode *)node->next;
		}
    }
}

bool CPtlPortal::RayCollisionWithDoor(CVector3D &p_start, CVector3D &p_directionUnit, float &p_distanceOut, CSurface3D **p_collisionSurfaceOut, CVector3D &p_collisionPointOut, CPtlDoor **p_collisionDoorOut)
{
    CPtlDoorListNode *node;
    static CVector3D newStart;
    static CVector3D newDirection;
	bool result = false;
    
    if (!doorList.IsEmpty())
	{
        node = (CPtlDoorListNode *)doorList.GetFirst();
		if (node)
        while (node->next != NULL)
		{
            newStart = node->currentOrient.TransformVectorToObjectSpace(p_start - node->currentOrient.p);
            newDirection = node->currentOrient.TransformVectorToObjectSpace(p_directionUnit);
            
            if (RayCollision(node->modelRef->GetSurfaces(), node->modelRef->GetSurfaceQty(), node->modelRef->GetVertices(), newStart, newDirection, p_distanceOut, p_collisionPointOut, p_collisionSurfaceOut))
			{
                result = true;

                *p_collisionDoorOut = node;
			}
            
            node = (CPtlDoorListNode *)node->next;
		}
	}

	return result;
}

bool CPtlPortal::SphereCollisionWithDoors(CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CVector3D &p_collisionNormalOut, CSurface3D **p_collisionSurfaceOut, CPtlDoor **p_collisionDoorOut)
{
    CPtlDoorListNode *node;
    static CVector3D newStart;
    static CVector3D newDirection;
	bool result = false;
    
    if (!doorList.IsEmpty())
	{
        node = (CPtlDoorListNode *)doorList.GetFirst();
		if (node)
        while (node->next != NULL)
		{
			if (node->IsClosed())
			{
				newStart = node->currentOrient.TransformVectorToObjectSpace(p_pos - node->currentOrient.p);
				newDirection = node->currentOrient.TransformVectorToObjectSpace(p_offset);
            
				if (SphereCollision(node->modelRef->GetSurfaces(), node->modelRef->GetSurfaceQty(), node->modelRef->GetVertices(), node->modelRef->GetVertexQty(), newStart, newDirection, p_radius, p_timeOut, p_collisionNormalOut, p_collisionSurfaceOut))
				{
					result = true;

					// Transform collision normal back to space of p_pos and p_offset
					p_collisionNormalOut =  node->currentOrient.TransformOffset(p_collisionNormalOut);

					*p_collisionDoorOut = node;
				}
			}
            
            node = (CPtlDoorListNode *)node->next;
		}
	}

	return result;
}

CPtlDoorListNode * CPtlPortal::AddDoor(void)
{
	return (CPtlDoorListNode *)doorList.NewNode();
}

