// ptldoor.cpp: implementation of the CPtlDoor class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PtlDoor.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

unsigned int CPtlDoor::currentCollisionID = 0;

CPtlDoor::CPtlDoor()
{
    // Assume a door opening at a 90 degree angle
    closedValue = 0.0f;
    openValue = 90.0f;
    speedMS = 90.0f / 1000.0f;
    
	// Starts closed
    targetValue = closedValue;
    currentValue = closedValue;
        
    modelRef = NULL;
	lightmaps = NULL;
	decalLists = NULL;

	renderID = 0;
}

CPtlDoor::~CPtlDoor()
{
	// Destroy lightmaps, they are specific to this door
	// Same for the decal lists

	Destroy();
}

void CPtlDoor::Destroy(void)
{
	if (decalLists)
	{
		delete [] decalLists;
		decalLists = NULL;
	}

	if (modelRef)
	if (lightmaps)
	{
		unsigned i;
		for (i=0; i<modelRef->GetSurfaceQty(); i++)
		{
			APITextures.DeleteTexture(lightmaps[i]);
		}

		delete [] lightmaps;
		lightmaps = NULL;
	}

	cellsTouched.EmptyList();
}

void CPtlDoor::Render(IRender *p_render)
{
	if (modelRef != NULL)
	{
		// Assemble a light list for door to use
		// Collect all light sources touching the same cells the door is touching
		CLightSourceRefList lights;
		CLightSourceRefListNode *lnode;
		CLinkedListNode <COctCellRef> *cnode1;
		CLinkedListNode <CLPVoid> *vnode1;

		lights.EmptyList();
		CLightSource::currentRenderID++;

		cnode1 = cellsTouched.GetFirst();
		if (cnode1)
		while (cnode1->next != NULL)
		{
			vnode1 = cnode1->cellRef->lpVoidList2.GetFirst();
			if (vnode1)
			while (vnode1->next != NULL)
			{
				if (((CLightSource *)vnode1->lpVoid)->renderID != CLightSource::currentRenderID)
				{
					((CLightSource *)vnode1->lpVoid)->renderID = CLightSource::currentRenderID;
					lnode = lights.NewNode();
					lnode->ref = (CLightSource *)(vnode1->lpVoid);
				}

				vnode1 = vnode1->next;
			}

			cnode1 = cnode1->next;
		}

		p_render->PushMatrix();
		p_render->TransformWorld(currentOrient);
		p_render->RenderModelFinalLM(modelRef, NULL, 255, lightmaps, decalLists, &lights, &currentOrient);
		p_render->PopMatrix();
	}
}

void CPtlDoor::Animate(long p_elapsedTimeMS)
{
    if (targetValue > currentValue)
	{
        currentValue = currentValue + speedMS * p_elapsedTimeMS;
        if (currentValue > targetValue)
            currentValue = targetValue;

        currentOrient = baseOrient;
        currentOrient.RotateV(currentValue, currentOrient.u);
	}
    else
	{
        if (targetValue < currentValue)
		{
            currentValue = currentValue - speedMS * p_elapsedTimeMS;
            if (currentValue < targetValue)
                currentValue = targetValue;

            currentOrient = baseOrient;
            currentOrient.RotateV(currentValue, currentOrient.u);
		}
	}
}

// For typical doors that are basically reshaped cubes with 6 sides
// Call after model set
// NOTE: The lightmap and Decallist IDs coudl already have been set on
//   this model from a previous call to this routine.  As long as
//   the ID assignments are always the same, it should be ok.
void CPtlDoor::SetupLists(CColorRGB p_lightmapColor, float p_texelPerUnit)
{
	if (!decalLists)
		decalLists = new CSurfaceDecalList[modelRef->GetSurfaceQty()];
	if (!lightmaps)
		lightmaps = new LPAPITEXTURE[modelRef->GetSurfaceQty()];

    unsigned i;
	unsigned dynLMID = 0;
	CPixelStorageMgr pixelStorageMgr;
    for (i=0; i<modelRef->GetSurfaceQty(); i++)
	{
        lightmaps[i] = APITextures.CreateNewTexture();
        
        modelRef->GetSurface(i).SetLightmapID(i);
        modelRef->GetSurface(i).SetDecalListID(i);
        
        modelRef->GetSurface(i).SetupLightmap(modelRef->GetVertices(), lightmaps, p_texelPerUnit, &pixelStorageMgr, &dynLMID);
        modelRef->GetSurface(i).ClearLightmap(p_lightmapColor, lightmaps);
    }
}

void CPtlDoor::DetermineCellsTouched(COctree &p_octree)
{
	if (!modelRef) return;

	CVector3D minPoint;
	CVector3D maxPoint;

	COctCellRefList cellList;

	float radius = modelRef->GetRadius();

	minPoint.x = currentOrient.p.x - radius;
	minPoint.y = currentOrient.p.y - radius;
	minPoint.z = currentOrient.p.z - radius;
	maxPoint.x = currentOrient.p.x + radius;
	maxPoint.y = currentOrient.p.y + radius;
	maxPoint.z = currentOrient.p.z + radius;

	p_octree.CellsTouched(minPoint.x, minPoint.y,minPoint.z, maxPoint.x, maxPoint.y, maxPoint.z, cellList);

	// Update cell object lists
	CLinkedListNode<COctCellRef> *cnode1, *cnode2;
	CLinkedListNode<CLPVoid> *vnode1;
	bool found;

	// For every cell in the old list that is not in the new list,
	//   remove the object from those cells
	cnode1 = cellsTouched.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		cnode2 = cellList.GetFirst();
		found = false;

		if (cnode2)
		while ((cnode2->next != NULL) && (!found))
		{
			if (cnode1->cellRef == cnode2->cellRef)
				found = true;
			else
				cnode2 = cnode2->next;
		}

		// Not there, remove this object from that cell's list
		if (!found)
		{
			vnode1 = cnode1->cellRef->lpVoidList1.GetFirst();
			found = false;
			if (vnode1)
			while ((vnode1->next != NULL) && (!found))
			{
				if (vnode1->lpVoid == this)
				{
					found = true;
					cnode1->cellRef->lpVoidList1.Delete(vnode1);
				}
				else
					vnode1 = vnode1->next;
			}
		}

		cnode1 = cnode1->next;
	}

	// For every cell in the new list that is not in the old list,
	//   add the object to those cells
	cnode1 = cellList.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		cnode2 = cellsTouched.GetFirst();
		found = false;

		if (cnode2)
		while ((cnode2->next != NULL) && (!found))
		{
			if (cnode1->cellRef == cnode2->cellRef)
				found = true;
			else
				cnode2 = cnode2->next;
		}

		// Not there, add this object to that cell's list
		if (!found)
		{
			cnode1->cellRef->lpVoidList1.NewNode()->lpVoid = this;
		}

		cnode1 = cnode1->next;
	}

	// Finally, replace the object's old list with the new list
	cellsTouched.EmptyList();
	if (!cellList.IsEmpty())
	{
		cellsTouched.GetHeader()->next = cellList.GetHeader()->next;
		cellsTouched.GetFooter()->prev = cellList.GetFooter()->prev;
		cellList.GetFirst()->prev = cellsTouched.GetHeader();
		cellList.GetLast()->next = cellsTouched.GetFooter();
		cellList.GetHeader()->next = cellList.GetFooter();
		cellList.GetFooter()->prev = cellList.GetHeader();
	}
}
