// LinkedList.h: interface for the CLinkedList class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_LINKEDLIST_H__1865FE6E_02D6_464C_A6DC_F30597AF3778__INCLUDED_)
#define AFX_LINKEDLIST_H__1865FE6E_02D6_464C_A6DC_F30597AF3778__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

template <class dtype>
class CLinkedListNode : public dtype	// Doing it this way merges
										//   all the members together at the same level
										//   and avoids the data variable
										// However, it won't allow
										//   int, unsigned lists, etc.
{
public:
	CLinkedListNode<dtype> *prev;
	CLinkedListNode<dtype> *next;
//		dtype data;

	// Don't use destructors here - can cause a stack overflow

	// Don't call this if you aren't sure the node is part of a list!
	inline void Detach(void)
	{
		this->prev->next = this->next;
		this->next->prev = this->prev;
	}
};

// Linked list class
// header and footer nodes are not valid data
// To parse, call GetFirst(), (if return NULL, list is empty)
//    proceed forward using node->next,
//    and if the current node == &list.footer (or node->next == NULL), you are done.
// You can also parse last to first, with GetLast(), and proceed using node->prev,
//    stopping at &list.header or node->prev == NULL
template <class dtype>
class CLinkedList
{
public:
	// NOTE: I prefer to keep header and footer as static nodes
	//   to reduce dynamic memory deletion when procedures are
	//   entered and left.  Typically, when a list is populated
	//   and finished, it is emptied (moved to the deleted list),
	//   and the deleted list is only destroyed on program exit
	CLinkedListNode<dtype> header;	// Start of the list
	CLinkedListNode<dtype> footer;	// End of the list

	// Make sure this is declared somewhere before using the list object
	static CLinkedList<dtype> DeletedList;

	// Make a header and footer
	CLinkedList()
	{
		header.prev = NULL;
		header.next = &footer;
		
		footer.prev = &header;
		footer.next = NULL;
	}

	// Destroy the list
	~CLinkedList()
	{
		CLinkedListNode<dtype> *temp1, *temp2;
		temp1 = header.next;

		if (&DeletedList.header == &header)
		{
			//TRACE("Killing a deleted list\n");
			while (temp1 != &footer)
			{
				temp2 = temp1->next;
				delete temp1;
				temp1 = temp2;
				TRACE("node gone\n");
			}
		}
		else
		{
			//TRACE("Killing (emptying) a main list\n");
			EmptyList();
		}
	}

	// Add a node to the list
	void Add (CLinkedListNode<dtype> *p_node)
	{
		p_node->next = header.next;
		p_node->prev = &header;
		header.next->prev = p_node;
		header.next = p_node;
	}

	// Same as Add
	inline void AddFirst (CLinkedListNode<dtype> *p_node)
	{
		Add (p_node);
	}

	// Add at end
	inline void Append (CLinkedListNode<dtype> *p_node)
	{
		p_node->next = &footer;
		p_node->prev = footer.prev;
		footer.prev->next = p_node;
		footer.prev = p_node;
	}
	inline void AddLast (CLinkedListNode<dtype> *p_node)
	{
		Append (p_node);
	}

	// Detach a node from the list
	void Detach (CLinkedListNode<dtype> *p_node)
	{
		p_node->prev->next = p_node->next;
		p_node->next->prev = p_node->prev;
	}

	// Delete a node from the list
	void Delete (CLinkedListNode<dtype> *p_node)
	{
		//TRACE("Moving a node to the deleted list\n");
		Detach (p_node);
		// Add node to the deleted list
		DeletedList.Add(p_node);
	}

	// Return the first valid parse node
	CLinkedListNode<dtype> * GetFirst(void)
	{
		if (header.next == &footer)
			// List empty
			return NULL;
		else
			return header.next;
	}
	
	// Return the last valid parse node
	CLinkedListNode<dtype> * GetLast(void)
	{
		if (header.next == &footer)
			// List empty
			return NULL;
		else
			return footer.prev;
	}

	// Create a new node
	// I would have preferred that this be a function of the GEListNode class,
	//   since it would make more sense, but the needed reference to the DeleteList
	//   object didn't make it possible - at least, I haven't figured that part out
	//   yet.
	CLinkedListNode<dtype> * NewNode (void)
	{
		return NewNode(true);
	}
	CLinkedListNode<dtype> * NewNode (bool p_append)
	{
		// Retrieve a node from the deleted list, otherwise create a new one

		// Look for node at the top of the deleted list
		// If found, detach it and return it
		CLinkedListNode<dtype> *temp;
		temp = DeletedList.GetFirst();

		if (temp == NULL)
		{
			// If none there, return a new one
			temp = new CLinkedListNode<dtype>;
		}
		else
		{
			DeletedList.Detach(temp);
			//TRACE("Getting a node from the deleted list\n");
		}

		// Add it to the list
		// Why would we not?
		// Well, to choose where we add it (beginning or end), for one
		if (p_append)
		{
			temp->next = &footer;
			temp->prev = footer.prev;
			footer.prev->next = temp;
			footer.prev = temp;
		}

		return temp;
	}
	static CLinkedListNode<dtype> * NewNodeDetached (void)
	{
		// Retrieve a node from the deleted list, otherwise create a new one

		// Look for node at the top of the deleted list
		// If found, detach it and return it
		CLinkedListNode<dtype> *temp;
		temp = DeletedList.GetFirst();

		if (temp == NULL)
		{
			// If none there, return a new one
			temp = new CLinkedListNode<dtype>;
		}
		else
		{
			DeletedList.Detach(temp);
			//TRACE("Getting a node from the deleted list\n");
		}

		return temp;
	}

	unsigned int GetNodeCount(void)
	{
		unsigned int count = 0;
		CLinkedListNode<dtype> *temp;

		temp = GetFirst();
		if (temp != NULL)
		while ( temp != &footer )
		{
			count++;
			temp = temp->next;
		}

		return count;
	}

	// Move all live content to deleted list
	void EmptyList(void)
	{
		// The second condition is for when we are getting
		//   rid of a list that has EmptyList() in its destructor
		// When it comes time to destroy DeletedList at the end of the
		//   program, we don't want a bunch of orphaned nodes somewhere
		// Just in case
		if ((GetFirst() != NULL ) && (&header != &DeletedList.header))
		{
			// Attach deleted list's footer to end of live list
			DeletedList.footer.prev->next = header.next;
			header.next->prev = DeletedList.footer.prev;

			// Attach first live node to end of deleted list (if any)
			DeletedList.footer.prev = footer.prev;
			footer.prev->next = &DeletedList.footer;

			// Attach live list's end nodes to each other
			footer.prev = &header;
			header.next = &footer;
		}
	}

	bool IsEmpty(void)
	{
		return (header.next == &footer);
	}

	CLinkedListNode<dtype> * GetHeader(void) {return &header;}
	CLinkedListNode<dtype> * GetFooter(void) {return &footer;}
};

// So we can make lists of integers
class CInteger
{
public:
	int value;
};

// So we can make lists of void pointers
class CLPVoid
{
public:
	void *lpVoid;
};

#endif // !defined(AFX_LINKEDLIST_H__1865FE6E_02D6_464C_A6DC_F30597AF3778__INCLUDED_)
